# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Basic implementation of a cache for general bots and gateway apps."""

from __future__ import annotations

__all__: typing.Sequence[str] = ("CacheImpl",)

import copy
import logging
import typing

from hikari import channels as channels_
from hikari import emojis
from hikari import messages
from hikari import snowflakes
from hikari import undefined
from hikari.api import cache
from hikari.api import config as config_api
from hikari.internal import cache as cache_utility
from hikari.internal import collections
from hikari.internal import typing_extensions

if typing.TYPE_CHECKING:
    from hikari import guilds
    from hikari import invites
    from hikari import presences
    from hikari import stickers
    from hikari import traits
    from hikari import users
    from hikari import voices
    from hikari.impl import config as config_impl

_LOGGER: typing.Final[logging.Logger] = logging.getLogger("hikari.cache")


# TODO: do we want to hide entities that are marked as "deleted" and being kept alive by references?
class CacheImpl(cache.MutableCache):
    """In-memory cache implementation.

    Parameters
    ----------
    app
        The object of the REST aware app this is bound to.
    settings
        The cache settings to use.
    """

    __slots__: typing.Sequence[str] = (
        "_app",
        "_dm_channel_entries",
        "_emoji_entries",
        "_guild_channel_entries",
        "_guild_entries",
        "_guild_thread_entries",
        "_intents",
        "_invite_entries",
        "_me",
        "_message_entries",
        "_referenced_messages",
        "_role_entries",
        "_settings",
        "_sticker_entries",
        "_unknown_custom_emoji_entries",
        "_user_entries",
    )

    # For the sake of keeping things clean, the annotations are being kept separate from the assignment here.
    _me: users.OwnUser | None
    _emoji_entries: collections.ExtendedMutableMapping[snowflakes.Snowflake, cache_utility.KnownCustomEmojiData]
    _dm_channel_entries: collections.ExtendedMutableMapping[snowflakes.Snowflake, snowflakes.Snowflake]
    _guild_channel_entries: collections.ExtendedMutableMapping[snowflakes.Snowflake, channels_.PermissibleGuildChannel]
    _guild_thread_entries: collections.ExtendedMutableMapping[snowflakes.Snowflake, channels_.GuildThreadChannel]
    _guild_entries: collections.ExtendedMutableMapping[snowflakes.Snowflake, cache_utility.GuildRecord]
    _invite_entries: collections.ExtendedMutableMapping[str, cache_utility.InviteData]
    _role_entries: collections.ExtendedMutableMapping[snowflakes.Snowflake, guilds.Role]
    _sticker_entries: collections.ExtendedMutableMapping[snowflakes.Snowflake, cache_utility.GuildStickerData]
    _unknown_custom_emoji_entries: collections.ExtendedMutableMapping[
        snowflakes.Snowflake, cache_utility.RefCell[emojis.CustomEmoji]
    ]
    _user_entries: collections.ExtendedMutableMapping[snowflakes.Snowflake, cache_utility.RefCell[users.User]]
    _message_entries: collections.ExtendedMutableMapping[
        snowflakes.Snowflake, cache_utility.RefCell[cache_utility.MessageData]
    ]
    _referenced_messages: collections.ExtendedMutableMapping[
        snowflakes.Snowflake, cache_utility.RefCell[cache_utility.MessageData]
    ]

    def __init__(self, app: traits.RESTAware, settings: config_impl.CacheSettings) -> None:
        self._app = app
        self._settings = settings
        self._create_cache()

    @property
    @typing_extensions.override
    def settings(self) -> config_impl.CacheSettings:
        return self._settings

    def _create_cache(self) -> None:
        self._me = None
        self._dm_channel_entries = collections.LimitedCapacityCacheMap(limit=self._settings.max_dm_channel_ids)
        self._emoji_entries = collections.FreezableDict()
        self._guild_channel_entries = collections.FreezableDict()
        self._guild_thread_entries = collections.FreezableDict()
        self._guild_entries = collections.FreezableDict()
        self._invite_entries = collections.FreezableDict()
        self._role_entries = collections.FreezableDict()
        self._sticker_entries = collections.FreezableDict()
        # This is a purely internal cache used for handling the caching and de-duplicating of the unknown custom emojis
        # found attached to cached presence activities.
        self._unknown_custom_emoji_entries = collections.FreezableDict()
        self._user_entries = collections.FreezableDict()
        self._message_entries = collections.LimitedCapacityCacheMap(
            limit=self._settings.max_messages, on_expire=self._on_message_expire
        )
        self._referenced_messages = collections.FreezableDict()

    def _is_cache_enabled_for(self, required_flag: config_api.CacheComponents) -> bool:
        return (self._settings.components & required_flag) == required_flag

    @staticmethod
    def _increment_ref_count(obj: cache_utility.RefCell[typing.Any], increment: int = 1) -> None:
        obj.ref_count += increment

    @typing_extensions.override
    def clear(self) -> None:
        if self._settings.components == config_api.CacheComponents.NONE:
            return

        self._create_cache()

    @typing_extensions.override
    def clear_dm_channel_ids(self) -> cache.CacheView[snowflakes.Snowflake, snowflakes.Snowflake]:
        if not self._is_cache_enabled_for(config_api.CacheComponents.DM_CHANNEL_IDS):
            return cache_utility.EmptyCacheView()

        result = self._dm_channel_entries
        self._dm_channel_entries = collections.LimitedCapacityCacheMap(limit=self._settings.max_dm_channel_ids)
        return cache_utility.CacheMappingView(result)

    @typing_extensions.override
    def delete_dm_channel_id(
        self, user: snowflakes.SnowflakeishOr[users.PartialUser], /
    ) -> snowflakes.Snowflake | None:
        if not self._is_cache_enabled_for(config_api.CacheComponents.DM_CHANNEL_IDS):
            return None

        return self._dm_channel_entries.pop(snowflakes.Snowflake(user), None)

    @typing_extensions.override
    def get_dm_channel_id(self, user: snowflakes.SnowflakeishOr[users.PartialUser], /) -> snowflakes.Snowflake | None:
        if not self._is_cache_enabled_for(config_api.CacheComponents.DM_CHANNEL_IDS):
            return None

        return self._dm_channel_entries.get(snowflakes.Snowflake(user))

    @typing_extensions.override
    def get_dm_channel_ids_view(self) -> cache.CacheView[snowflakes.Snowflake, snowflakes.Snowflake]:
        if not self._is_cache_enabled_for(config_api.CacheComponents.DM_CHANNEL_IDS):
            return cache_utility.EmptyCacheView()

        return cache_utility.CacheMappingView(self._dm_channel_entries.freeze())

    @typing_extensions.override
    def set_dm_channel_id(
        self,
        user: snowflakes.SnowflakeishOr[users.PartialUser],
        channel: snowflakes.SnowflakeishOr[channels_.PartialChannel],
        /,
    ) -> None:
        if not self._is_cache_enabled_for(config_api.CacheComponents.DM_CHANNEL_IDS):
            return

        self._dm_channel_entries[snowflakes.Snowflake(user)] = snowflakes.Snowflake(channel)

    def _build_emoji(self, emoji_data: cache_utility.KnownCustomEmojiData) -> emojis.KnownCustomEmoji:
        return emoji_data.build_entity(self._app)

    @typing_extensions.override
    def clear_emojis(self) -> cache.CacheView[snowflakes.Snowflake, emojis.KnownCustomEmoji]:
        if not self._is_cache_enabled_for(config_api.CacheComponents.EMOJIS):
            return cache_utility.EmptyCacheView()

        cached_emojis = self._emoji_entries
        self._emoji_entries = collections.FreezableDict()

        for emoji_data in cached_emojis.values():
            if emoji_data.user:
                self._garbage_collect_user(emoji_data.user, decrement=1)

        for guild_id, guild_record in self._guild_entries.freeze().items():
            guild_record.emojis = None
            self._remove_guild_record_if_empty(guild_id, guild_record)

        return cache_utility.CacheMappingView(cached_emojis, builder=self._build_emoji)

    @typing_extensions.override
    def clear_emojis_for_guild(
        self, guild: snowflakes.SnowflakeishOr[guilds.PartialGuild], /
    ) -> cache.CacheView[snowflakes.Snowflake, emojis.KnownCustomEmoji]:
        if not self._is_cache_enabled_for(config_api.CacheComponents.EMOJIS):
            return cache_utility.EmptyCacheView()

        guild_id = snowflakes.Snowflake(guild)
        guild_record = self._guild_entries.get(guild_id)
        if not guild_record or not guild_record.emojis:
            return cache_utility.EmptyCacheView()

        cached_emojis = {emoji_id: self._emoji_entries.pop(emoji_id) for emoji_id in guild_record.emojis}
        guild_record.emojis = None
        self._remove_guild_record_if_empty(guild_id, guild_record)

        for emoji_data in cached_emojis.values():
            if emoji_data.user:
                self._garbage_collect_user(emoji_data.user, decrement=1)

        return cache_utility.CacheMappingView(cached_emojis, builder=self._build_emoji)

    @typing_extensions.override
    def delete_emoji(self, emoji: snowflakes.SnowflakeishOr[emojis.CustomEmoji], /) -> emojis.KnownCustomEmoji | None:
        if not self._is_cache_enabled_for(config_api.CacheComponents.EMOJIS):
            return None

        emoji_id = snowflakes.Snowflake(emoji)
        emoji_data = self._emoji_entries.pop(emoji_id, None)
        if not emoji_data:
            return None

        if emoji_data.user:
            self._garbage_collect_user(emoji_data.user, decrement=1)

        guild_record = self._guild_entries.get(emoji_data.guild_id)
        if guild_record and guild_record.emojis:
            guild_record.emojis.remove(emoji_id)

            if not guild_record.emojis:
                guild_record.emojis = None
                self._remove_guild_record_if_empty(emoji_data.guild_id, guild_record)

        return self._build_emoji(emoji_data)

    @typing_extensions.override
    def get_emoji(self, emoji: snowflakes.SnowflakeishOr[emojis.CustomEmoji], /) -> emojis.KnownCustomEmoji | None:
        if not self._is_cache_enabled_for(config_api.CacheComponents.EMOJIS):
            return None

        emoji_data = self._emoji_entries.get(snowflakes.Snowflake(emoji))
        return self._build_emoji(emoji_data) if emoji_data else None

    @typing_extensions.override
    def get_emojis_view(self) -> cache.CacheView[snowflakes.Snowflake, emojis.KnownCustomEmoji]:
        if not self._is_cache_enabled_for(config_api.CacheComponents.EMOJIS):
            return cache_utility.EmptyCacheView()

        return cache_utility.CacheMappingView(self._emoji_entries.freeze(), builder=self._build_emoji)

    @typing_extensions.override
    def get_emojis_view_for_guild(
        self, guild: snowflakes.SnowflakeishOr[guilds.PartialGuild], /
    ) -> cache.CacheView[snowflakes.Snowflake, emojis.KnownCustomEmoji]:
        if not self._is_cache_enabled_for(config_api.CacheComponents.EMOJIS):
            return cache_utility.EmptyCacheView()

        guild_record = self._guild_entries.get(snowflakes.Snowflake(guild))
        if not guild_record or not guild_record.emojis:
            return cache_utility.EmptyCacheView()

        cached_emojis = {emoji_id: self._emoji_entries[emoji_id] for emoji_id in guild_record.emojis}
        return cache_utility.CacheMappingView(cached_emojis, builder=self._build_emoji)

    @typing_extensions.override
    def set_emoji(self, emoji: emojis.KnownCustomEmoji, /) -> None:
        if not self._is_cache_enabled_for(config_api.CacheComponents.EMOJIS):
            return

        if not emoji.guild_id:
            msg = "Cannot cache an emoji without a guild ID."
            raise ValueError(msg)

        user: cache_utility.RefCell[users.User] | None = None
        if emoji.user:
            user = self._set_user(emoji.user)
            if emoji.id not in self._emoji_entries:
                self._increment_ref_count(user)

        emoji_data = cache_utility.KnownCustomEmojiData.build_from_entity(emoji, user=user)
        self._emoji_entries[emoji.id] = emoji_data
        guild_record = self._get_or_create_guild_record(emoji.guild_id)

        if guild_record.emojis is None:  # TODO: add test cases when it is not None?
            guild_record.emojis = collections.SnowflakeSet()

        guild_record.emojis.add(emoji.id)

    @typing_extensions.override
    def update_emoji(
        self, emoji: emojis.KnownCustomEmoji, /
    ) -> tuple[emojis.KnownCustomEmoji | None, emojis.KnownCustomEmoji | None]:
        if not self._is_cache_enabled_for(config_api.CacheComponents.EMOJIS):
            return None, None

        if not emoji.guild_id:
            msg = "Emoji must have a guild ID to be cached."
            raise ValueError(msg)

        cached_emoji = self.get_emoji(emoji.id)
        self.set_emoji(emoji)
        return cached_emoji, self.get_emoji(emoji.id)

    def _build_sticker(self, sticker_data: cache_utility.GuildStickerData) -> stickers.GuildSticker:
        return sticker_data.build_entity(self._app)

    @typing_extensions.override
    def clear_stickers(self) -> cache.CacheView[snowflakes.Snowflake, stickers.GuildSticker]:
        if not self._is_cache_enabled_for(config_api.CacheComponents.GUILD_STICKERS):
            return cache_utility.EmptyCacheView()

        cached_stickers = self._sticker_entries
        self._sticker_entries = collections.FreezableDict()

        for sticker_data in cached_stickers.values():
            if sticker_data.user:
                self._garbage_collect_user(sticker_data.user, decrement=1)

        for guild_id, guild_record in self._guild_entries.freeze().items():
            guild_record.stickers = None
            self._remove_guild_record_if_empty(guild_id, guild_record)

        return cache_utility.CacheMappingView(cached_stickers, builder=self._build_sticker)

    @typing_extensions.override
    def clear_stickers_for_guild(
        self, guild: snowflakes.SnowflakeishOr[guilds.PartialGuild], /
    ) -> cache.CacheView[snowflakes.Snowflake, stickers.GuildSticker]:
        if not self._is_cache_enabled_for(config_api.CacheComponents.GUILD_STICKERS):
            return cache_utility.EmptyCacheView()

        guild_id = snowflakes.Snowflake(guild)
        guild_record = self._guild_entries.get(guild_id)
        if not guild_record or not guild_record.stickers:
            return cache_utility.EmptyCacheView()

        cached_stickers = {sticker_id: self._sticker_entries.pop(sticker_id) for sticker_id in guild_record.stickers}
        guild_record.stickers = None
        self._remove_guild_record_if_empty(guild_id, guild_record)

        for sticker_data in cached_stickers.values():
            if sticker_data.user:
                self._garbage_collect_user(sticker_data.user, decrement=1)

        return cache_utility.CacheMappingView(cached_stickers, builder=self._build_sticker)

    @typing_extensions.override
    def get_sticker(self, sticker: snowflakes.SnowflakeishOr[stickers.GuildSticker], /) -> stickers.GuildSticker | None:
        if not self._is_cache_enabled_for(config_api.CacheComponents.GUILD_STICKERS):
            return None

        sticker_data = self._sticker_entries.get(snowflakes.Snowflake(sticker))
        return self._build_sticker(sticker_data) if sticker_data else None

    @typing_extensions.override
    def get_stickers_view(self) -> cache.CacheView[snowflakes.Snowflake, stickers.GuildSticker]:
        if not self._is_cache_enabled_for(config_api.CacheComponents.GUILD_STICKERS):
            return cache_utility.EmptyCacheView()

        return cache_utility.CacheMappingView(self._sticker_entries.freeze(), builder=self._build_sticker)

    @typing_extensions.override
    def get_stickers_view_for_guild(
        self, guild: snowflakes.SnowflakeishOr[guilds.PartialGuild], /
    ) -> cache.CacheView[snowflakes.Snowflake, stickers.GuildSticker]:
        if not self._is_cache_enabled_for(config_api.CacheComponents.GUILD_STICKERS):
            return cache_utility.EmptyCacheView()

        guild_record = self._guild_entries.get(snowflakes.Snowflake(guild))
        if not guild_record or not guild_record.stickers:
            return cache_utility.EmptyCacheView()

        cached_stickers = {sticker_id: self._sticker_entries[sticker_id] for sticker_id in guild_record.stickers}
        return cache_utility.CacheMappingView(cached_stickers, builder=self._build_sticker)

    @typing_extensions.override
    def delete_sticker(
        self, sticker: snowflakes.SnowflakeishOr[stickers.GuildSticker], /
    ) -> stickers.GuildSticker | None:
        if not self._is_cache_enabled_for(config_api.CacheComponents.GUILD_STICKERS):
            return None

        sticker_id = snowflakes.Snowflake(sticker)
        sticker_data = self._sticker_entries.pop(sticker_id, None)
        if not sticker_data:
            return None

        if sticker_data.user:
            self._garbage_collect_user(sticker_data.user, decrement=1)

        guild_record = self._guild_entries.get(sticker_data.guild_id)
        if guild_record and guild_record.stickers:
            guild_record.stickers.remove(sticker_id)

            if not guild_record.stickers:
                guild_record.stickers = None

        return self._build_sticker(sticker_data)

    @typing_extensions.override
    def set_sticker(self, sticker: stickers.GuildSticker, /) -> None:
        if not self._is_cache_enabled_for(config_api.CacheComponents.GUILD_STICKERS):
            return

        user: cache_utility.RefCell[users.User] | None = None
        if sticker.user:
            user = self._set_user(sticker.user)
            if sticker.id not in self._sticker_entries:
                self._increment_ref_count(user)

        sticker_data = cache_utility.GuildStickerData.build_from_entity(sticker, user=user)
        self._sticker_entries[sticker.id] = sticker_data
        guild_record = self._get_or_create_guild_record(sticker.guild_id)

        if guild_record.stickers is None:  # TODO: add test cases when it is not None?
            guild_record.stickers = collections.SnowflakeSet()

        guild_record.stickers.add(sticker.id)

    def _remove_guild_record_if_empty(
        self, guild_id: snowflakes.Snowflake, record: cache_utility.GuildRecord, /
    ) -> None:
        if guild_id in self._guild_entries and record.empty():
            del self._guild_entries[guild_id]

    def _get_or_create_guild_record(self, guild_id: snowflakes.Snowflake) -> cache_utility.GuildRecord:
        if guild_id not in self._guild_entries:
            self._guild_entries[guild_id] = cache_utility.GuildRecord()

        return self._guild_entries[guild_id]

    @typing_extensions.override
    def clear_guilds(self) -> cache.CacheView[snowflakes.Snowflake, guilds.GatewayGuild]:
        if not self._is_cache_enabled_for(config_api.CacheComponents.GUILDS):
            return cache_utility.EmptyCacheView()

        cached_guilds = {}

        for guild_id, guild_record in self._guild_entries.freeze().items():
            if guild_record.guild:
                cached_guilds[guild_id] = guild_record.guild
                guild_record.guild = None
                guild_record.is_available = None
                self._remove_guild_record_if_empty(guild_id, guild_record)

        return cache_utility.CacheMappingView(cached_guilds) if cached_guilds else cache_utility.EmptyCacheView()

    @typing_extensions.override
    def delete_guild(self, guild: snowflakes.SnowflakeishOr[guilds.PartialGuild], /) -> guilds.GatewayGuild | None:
        if not self._is_cache_enabled_for(config_api.CacheComponents.GUILDS):
            return None

        guild_id = snowflakes.Snowflake(guild)
        guild_record = self._guild_entries.get(guild_id)
        if not guild_record:
            return None

        guild = guild_record.guild

        if guild:
            guild_record.guild = None
            guild_record.is_available = None
            self._remove_guild_record_if_empty(guild_id, guild_record)

        return guild

    def _get_guild(
        self, guild: snowflakes.SnowflakeishOr[guilds.PartialGuild], /, *, availability: bool
    ) -> guilds.GatewayGuild | None:
        guild_record = self._guild_entries.get(snowflakes.Snowflake(guild))
        if not guild_record or not guild_record.guild or guild_record.is_available is not availability:
            return None

        return copy.copy(guild_record.guild)

    @typing_extensions.override
    def get_guild(self, guild: snowflakes.SnowflakeishOr[guilds.PartialGuild], /) -> guilds.GatewayGuild | None:
        if not self._is_cache_enabled_for(config_api.CacheComponents.GUILDS):
            return None

        guild_record = self._guild_entries.get(snowflakes.Snowflake(guild))
        return copy.copy(guild_record.guild) if guild_record and guild_record.guild else None

    @typing_extensions.override
    def get_available_guild(
        self, guild: snowflakes.SnowflakeishOr[guilds.PartialGuild], /
    ) -> guilds.GatewayGuild | None:
        if not self._is_cache_enabled_for(config_api.CacheComponents.GUILDS):
            return None

        return self._get_guild(guild, availability=True)

    @typing_extensions.override
    def get_unavailable_guild(
        self, guild: snowflakes.SnowflakeishOr[guilds.PartialGuild], /
    ) -> guilds.GatewayGuild | None:
        if not self._is_cache_enabled_for(config_api.CacheComponents.GUILDS):
            return None

        return self._get_guild(guild, availability=False)

    def _get_guilds_view(self, *, availability: bool) -> cache.CacheView[snowflakes.Snowflake, guilds.GatewayGuild]:
        # We may have a guild record without a guild object in cases where we're caching other entities that belong to
        # the guild therefore we want to make sure record.guild isn't None.
        results = {
            sf: guild_record.guild
            for sf, guild_record in self._guild_entries.items()
            if guild_record.guild and guild_record.is_available is availability
        }
        return cache_utility.CacheMappingView(results) if results else cache_utility.EmptyCacheView()

    @typing_extensions.override
    def get_guilds_view(self) -> cache.CacheView[snowflakes.Snowflake, guilds.GatewayGuild]:
        return cache_utility.CacheMappingView(
            {guild_id: record.guild for guild_id, record in self._guild_entries.items() if record.guild}
        )

    @typing_extensions.override
    def get_available_guilds_view(self) -> cache.CacheView[snowflakes.Snowflake, guilds.GatewayGuild]:
        if not self._is_cache_enabled_for(config_api.CacheComponents.GUILDS):
            return cache_utility.EmptyCacheView()

        return self._get_guilds_view(availability=True)

    @typing_extensions.override
    def get_unavailable_guilds_view(self) -> cache.CacheView[snowflakes.Snowflake, guilds.GatewayGuild]:
        if not self._is_cache_enabled_for(config_api.CacheComponents.GUILDS):
            return cache_utility.EmptyCacheView()

        return self._get_guilds_view(availability=False)

    @typing_extensions.override
    def set_guild(self, guild: guilds.GatewayGuild, /) -> None:
        if not self._is_cache_enabled_for(config_api.CacheComponents.GUILDS):
            return

        guild_record = self._get_or_create_guild_record(guild.id)
        guild_record.guild = copy.copy(guild)
        guild_record.is_available = True

    @typing_extensions.override
    def set_guild_availability(
        self,
        guild: snowflakes.SnowflakeishOr[guilds.PartialGuild],
        is_available: bool,  # noqa: FBT001 - Boolean-typed positional argument
        /,
    ) -> None:
        if not self._is_cache_enabled_for(config_api.CacheComponents.GUILDS):
            return

        guild_record = self._guild_entries.get(snowflakes.Snowflake(guild))
        if guild_record and guild_record.guild:
            guild_record.is_available = is_available

    @typing_extensions.override
    def update_guild(
        self, guild: guilds.GatewayGuild, /
    ) -> tuple[guilds.GatewayGuild | None, guilds.GatewayGuild | None]:
        if not self._is_cache_enabled_for(config_api.CacheComponents.GUILDS):
            return None, None

        guild = copy.copy(guild)
        cached_guild = self.get_guild(guild.id)

        # We have to manually update these because Inconsistency is Discord's middle name.
        if cached_guild:
            guild.member_count = cached_guild.member_count
            guild.joined_at = cached_guild.joined_at
            guild.is_large = cached_guild.is_large

        self.set_guild(guild)
        return cached_guild, self.get_guild(guild.id)

    @typing_extensions.override
    def clear_threads(self) -> cache.CacheView[snowflakes.Snowflake, channels_.GuildThreadChannel]:
        if not self._is_cache_enabled_for(config_api.CacheComponents.GUILD_THREADS):
            return cache_utility.EmptyCacheView()

        cached_threads = self._guild_thread_entries
        self._guild_thread_entries = collections.FreezableDict()

        for guild_id, guild_record in self._guild_entries.freeze().items():
            if guild_record.threads:
                guild_record.threads = None
                self._remove_guild_record_if_empty(guild_id, guild_record)

        return cache_utility.CacheMappingView(cached_threads)

    @typing_extensions.override
    def clear_threads_for_guild(
        self, guild: snowflakes.SnowflakeishOr[guilds.PartialGuild], /
    ) -> cache.CacheView[snowflakes.Snowflake, channels_.GuildThreadChannel]:
        if not self._is_cache_enabled_for(config_api.CacheComponents.GUILD_THREADS):
            return cache_utility.EmptyCacheView()

        guild_id = snowflakes.Snowflake(guild)
        guild_record = self._guild_entries.get(guild_id)
        if not guild_record or not guild_record.threads:
            return cache_utility.EmptyCacheView()

        cached_threads = {sf: self._guild_thread_entries.pop(sf) for sf in guild_record.threads}
        guild_record.threads = None
        self._remove_guild_record_if_empty(guild_id, guild_record)
        return cache_utility.CacheMappingView(cached_threads)

    @typing_extensions.override
    def clear_threads_for_channel(
        self,
        guild: snowflakes.SnowflakeishOr[guilds.PartialGuild],
        channel: snowflakes.SnowflakeishOr[channels_.PartialChannel],
        /,
    ) -> cache.CacheView[snowflakes.Snowflake, channels_.GuildThreadChannel]:
        if not self._is_cache_enabled_for(config_api.CacheComponents.GUILD_THREADS):
            return cache_utility.EmptyCacheView()

        channel_id = snowflakes.Snowflake(channel)
        guild = snowflakes.Snowflake(guild)
        guild_record = self._guild_entries.get(guild)
        if not guild_record or not guild_record.threads:
            return cache_utility.EmptyCacheView()

        threads: dict[snowflakes.Snowflake, channels_.GuildThreadChannel] = {}
        for thread in map(self._guild_thread_entries.__getitem__, tuple(guild_record.threads)):
            if thread.parent_id == channel_id:
                del self._guild_thread_entries[thread.id]
                guild_record.threads.remove(thread.id)

        if not guild_record.threads:
            guild_record.threads = None
            self._remove_guild_record_if_empty(guild, guild_record)

        return cache_utility.CacheMappingView(threads)

    @typing_extensions.override
    def delete_thread(
        self, thread: snowflakes.SnowflakeishOr[channels_.PartialChannel], /
    ) -> channels_.GuildThreadChannel | None:
        if not self._is_cache_enabled_for(config_api.CacheComponents.GUILD_THREADS):
            return None

        thread_id = snowflakes.Snowflake(thread)
        thread = self._guild_thread_entries.pop(thread_id, None)

        if not thread:
            return None

        guild_record = self._guild_entries.get(thread.guild_id)
        if guild_record and guild_record.threads:
            guild_record.threads.remove(thread_id)
            if not guild_record.threads:
                guild_record.threads = None
                self._remove_guild_record_if_empty(thread.guild_id, guild_record)

        return thread

    @typing_extensions.override
    def get_thread(
        self, thread: snowflakes.SnowflakeishOr[channels_.PartialChannel], /
    ) -> channels_.GuildThreadChannel | None:
        if not self._is_cache_enabled_for(config_api.CacheComponents.GUILD_THREADS):
            return None

        thread = self._guild_thread_entries.get(snowflakes.Snowflake(thread))
        return copy.copy(thread) if thread else None

    @typing_extensions.override
    def get_threads_view(self) -> cache.CacheView[snowflakes.Snowflake, channels_.GuildThreadChannel]:
        return cache_utility.CacheMappingView(self._guild_thread_entries.freeze())

    @typing_extensions.override
    def get_threads_view_for_guild(
        self, guild: snowflakes.SnowflakeishOr[guilds.PartialGuild], /
    ) -> cache.CacheView[snowflakes.Snowflake, channels_.GuildThreadChannel]:
        if not self._is_cache_enabled_for(config_api.CacheComponents.GUILD_THREADS):
            return cache_utility.EmptyCacheView()

        guild_record = self._guild_entries.get(snowflakes.Snowflake(guild))
        if not guild_record or not guild_record.threads:
            return cache_utility.EmptyCacheView()

        return cache_utility.CacheMappingView({sf: self._guild_thread_entries[sf] for sf in guild_record.threads})

    @typing_extensions.override
    def get_threads_view_for_channel(
        self,
        guild: snowflakes.SnowflakeishOr[guilds.PartialGuild],
        channel: snowflakes.SnowflakeishOr[channels_.PartialChannel],
        /,
    ) -> cache.CacheView[snowflakes.Snowflake, channels_.GuildThreadChannel]:
        if not self._is_cache_enabled_for(config_api.CacheComponents.GUILD_THREADS):
            return cache_utility.EmptyCacheView()

        record = self._guild_entries.get(snowflakes.Snowflake(guild))
        if not record or not record.threads:
            return cache_utility.EmptyCacheView()

        threads = map(self._guild_thread_entries.__getitem__, record.threads)
        channel = snowflakes.Snowflake(channel)
        return cache_utility.CacheMappingView({thread.id: thread for thread in threads if thread.parent_id == channel})

    @typing_extensions.override
    def set_thread(self, thread: channels_.GuildThreadChannel, /) -> None:
        if not self._is_cache_enabled_for(config_api.CacheComponents.GUILD_THREADS):
            return

        self._guild_thread_entries[thread.id] = copy.copy(thread)
        guild_record = self._get_or_create_guild_record(thread.guild_id)

        if guild_record.threads is None:
            guild_record.threads = collections.SnowflakeSet()

        guild_record.threads.add(thread.id)

    @typing_extensions.override
    def update_thread(
        self, thread: channels_.GuildThreadChannel, /
    ) -> tuple[channels_.GuildThreadChannel | None, channels_.GuildThreadChannel | None]:
        if not self._is_cache_enabled_for(config_api.CacheComponents.GUILD_THREADS):
            return None, None

        cached_thread = self.get_thread(thread.id)
        self.set_thread(thread)
        return cached_thread, self.get_thread(thread.id)

    @typing_extensions.override
    def clear_guild_channels(self) -> cache.CacheView[snowflakes.Snowflake, channels_.PermissibleGuildChannel]:
        if not self._is_cache_enabled_for(config_api.CacheComponents.GUILD_CHANNELS):
            return cache_utility.EmptyCacheView()

        cached_channels = self._guild_channel_entries
        self._guild_channel_entries = collections.FreezableDict()

        for guild_id, guild_record in self._guild_entries.freeze().items():
            if guild_record.channels:
                guild_record.channels = None
                self._remove_guild_record_if_empty(guild_id, guild_record)

        return cache_utility.CacheMappingView(cached_channels)

    @typing_extensions.override
    def clear_guild_channels_for_guild(
        self, guild: snowflakes.SnowflakeishOr[guilds.PartialGuild], /
    ) -> cache.CacheView[snowflakes.Snowflake, channels_.PermissibleGuildChannel]:
        if not self._is_cache_enabled_for(config_api.CacheComponents.GUILD_CHANNELS):
            return cache_utility.EmptyCacheView()

        guild_id = snowflakes.Snowflake(guild)
        guild_record = self._guild_entries.get(guild_id)
        if not guild_record or not guild_record.channels:
            return cache_utility.EmptyCacheView()

        cached_channels = {sf: self._guild_channel_entries.pop(sf) for sf in guild_record.channels}
        guild_record.channels = None
        self._remove_guild_record_if_empty(guild_id, guild_record)
        return cache_utility.CacheMappingView(cached_channels)

    @typing_extensions.override
    def delete_guild_channel(
        self, channel: snowflakes.SnowflakeishOr[channels_.PartialChannel], /
    ) -> channels_.PermissibleGuildChannel | None:
        if not self._is_cache_enabled_for(config_api.CacheComponents.GUILD_CHANNELS):
            return None

        channel_id = snowflakes.Snowflake(channel)
        channel = self._guild_channel_entries.pop(channel_id, None)

        if not channel:
            return None

        guild_record = self._guild_entries.get(channel.guild_id)
        if guild_record and guild_record.channels:
            guild_record.channels.remove(channel_id)
            if not guild_record.channels:
                guild_record.channels = None
                self._remove_guild_record_if_empty(channel.guild_id, guild_record)

        return channel

    @typing_extensions.override
    def get_guild_channel(
        self, channel: snowflakes.SnowflakeishOr[channels_.PartialChannel], /
    ) -> channels_.PermissibleGuildChannel | None:
        if not self._is_cache_enabled_for(config_api.CacheComponents.GUILD_CHANNELS):
            return None

        channel = self._guild_channel_entries.get(snowflakes.Snowflake(channel))
        return cache_utility.copy_guild_channel(channel) if channel else None

    @typing_extensions.override
    def get_guild_channels_view(self) -> cache.CacheView[snowflakes.Snowflake, channels_.PermissibleGuildChannel]:
        if not self._is_cache_enabled_for(config_api.CacheComponents.GUILD_CHANNELS):
            return cache_utility.EmptyCacheView()

        return cache_utility.CacheMappingView(
            self._guild_channel_entries.freeze(),
            builder=cache_utility.copy_guild_channel,  # type: ignore[type-var]
        )

    @typing_extensions.override
    def get_guild_channels_view_for_guild(
        self, guild: snowflakes.SnowflakeishOr[guilds.PartialGuild], /
    ) -> cache.CacheView[snowflakes.Snowflake, channels_.PermissibleGuildChannel]:
        if not self._is_cache_enabled_for(config_api.CacheComponents.GUILD_CHANNELS):
            return cache_utility.EmptyCacheView()

        guild_record = self._guild_entries.get(snowflakes.Snowflake(guild))
        if not guild_record or not guild_record.channels:
            return cache_utility.EmptyCacheView()

        cached_channels = {sf: self._guild_channel_entries[sf] for sf in guild_record.channels}

        def sorter(args: tuple[snowflakes.Snowflake, channels_.PermissibleGuildChannel]) -> tuple[int, int, int]:
            channel = args[1]
            if isinstance(channel, channels_.GuildCategory):
                return channel.position, -1, 0

            parent_position = -1 if channel.parent_id is None else cached_channels[channel.parent_id].position

            if not isinstance(channel, channels_.GuildVoiceChannel):
                return parent_position, 0, channel.position

            return parent_position, 1, channel.position

        cached_channels = dict(sorted(cached_channels.items(), key=sorter))
        return cache_utility.CacheMappingView(
            cached_channels,
            builder=cache_utility.copy_guild_channel,  # type: ignore[type-var]
        )

    @typing_extensions.override
    def set_guild_channel(self, channel: channels_.PermissibleGuildChannel, /) -> None:
        if not self._is_cache_enabled_for(config_api.CacheComponents.GUILD_CHANNELS):
            return

        self._guild_channel_entries[channel.id] = cache_utility.copy_guild_channel(channel)
        guild_record = self._get_or_create_guild_record(channel.guild_id)

        if guild_record.channels is None:
            guild_record.channels = collections.SnowflakeSet()

        guild_record.channels.add(channel.id)

    @typing_extensions.override
    def update_guild_channel(
        self, channel: channels_.PermissibleGuildChannel, /
    ) -> tuple[channels_.PermissibleGuildChannel | None, channels_.PermissibleGuildChannel | None]:
        if not self._is_cache_enabled_for(config_api.CacheComponents.GUILD_CHANNELS):
            return None, None

        cached_channel = self.get_guild_channel(channel.id)
        self.set_guild_channel(channel)
        return cached_channel, self.get_guild_channel(channel.id)

    def _build_invite(self, invite_data: cache_utility.InviteData) -> invites.InviteWithMetadata:
        return invite_data.build_entity(self._app)

    def _remove_invite_users(self, invite: cache_utility.InviteData) -> None:
        if invite.inviter:
            self._garbage_collect_user(invite.inviter, decrement=1)

        if invite.target_user:
            self._garbage_collect_user(invite.target_user, decrement=1)

    @typing_extensions.override
    def clear_invites(self) -> cache.CacheView[str, invites.InviteWithMetadata]:
        if not self._is_cache_enabled_for(config_api.CacheComponents.INVITES):
            return cache_utility.EmptyCacheView()

        cached_invites = self._invite_entries
        self._invite_entries = collections.FreezableDict()

        for invite_data in cached_invites.values():
            self._remove_invite_users(invite_data)

        for guild_id, guild_record in self._guild_entries.freeze().items():
            if guild_record.invites:
                guild_record.invites = None
                self._remove_guild_record_if_empty(guild_id, guild_record)

        return cache_utility.CacheMappingView(cached_invites, builder=self._build_invite)

    @typing_extensions.override
    def clear_invites_for_guild(
        self, guild: snowflakes.SnowflakeishOr[guilds.PartialGuild], /
    ) -> cache.CacheView[str, invites.InviteWithMetadata]:
        if not self._is_cache_enabled_for(config_api.CacheComponents.INVITES):
            return cache_utility.EmptyCacheView()

        guild_id = snowflakes.Snowflake(guild)
        guild_record = self._guild_entries.get(guild_id)

        if not guild_record or not guild_record.invites:
            return cache_utility.EmptyCacheView()

        cached_invites = {invite_code: self._invite_entries.pop(invite_code) for invite_code in guild_record.invites}
        guild_record.invites = None
        self._remove_guild_record_if_empty(guild_id, guild_record)

        for invite_data in cached_invites.values():
            self._remove_invite_users(invite_data)

        return cache_utility.CacheMappingView(cached_invites, builder=self._build_invite)

    @typing_extensions.override
    def clear_invites_for_channel(
        self,
        guild: snowflakes.SnowflakeishOr[guilds.PartialGuild],
        channel: snowflakes.SnowflakeishOr[channels_.PartialChannel],
        /,
    ) -> cache.CacheView[str, invites.InviteWithMetadata]:
        if not self._is_cache_enabled_for(config_api.CacheComponents.INVITES):
            return cache_utility.EmptyCacheView()

        guild_id = snowflakes.Snowflake(guild)
        channel_id = snowflakes.Snowflake(channel)
        guild_record = self._guild_entries.get(guild_id)
        if not guild_record or not guild_record.invites:
            return cache_utility.EmptyCacheView()

        cached_invites = {}

        for code in tuple(guild_record.invites):
            invite_data = self._invite_entries[code]
            if invite_data.channel_id != channel_id:
                continue

            cached_invites[code] = invite_data
            del self._invite_entries[code]
            guild_record.invites.remove(code)
            self._remove_invite_users(invite_data)

        if not guild_record.invites:
            guild_record.invites = None
            self._remove_guild_record_if_empty(guild_id, guild_record)

        return cache_utility.CacheMappingView(cached_invites, builder=self._build_invite)

    @typing_extensions.override
    def delete_invite(self, code: invites.InviteCode | str, /) -> invites.InviteWithMetadata | None:
        if not self._is_cache_enabled_for(config_api.CacheComponents.INVITES):
            return None

        code = code if isinstance(code, str) else code.code
        invite_data = self._invite_entries.pop(code, None)
        if not invite_data:
            return None

        self._remove_invite_users(invite_data)

        if invite_data.guild_id is not None:  # TODO: test case when this is None?
            guild_record = self._guild_entries.get(invite_data.guild_id)
            if guild_record and guild_record.invites:
                guild_record.invites.remove(code)

                if not guild_record.invites:
                    guild_record.invites = None  # TODO: test when this is set to None
                    self._remove_guild_record_if_empty(invite_data.guild_id, guild_record)

        return self._build_invite(invite_data)

    @typing_extensions.override
    def get_invite(self, code: invites.InviteCode | str, /) -> invites.InviteWithMetadata | None:
        if not self._is_cache_enabled_for(config_api.CacheComponents.INVITES):
            return None

        code = code if isinstance(code, str) else code.code
        invite_data = self._invite_entries.get(code)
        return self._build_invite(invite_data) if invite_data else None

    @typing_extensions.override
    def get_invites_view(self) -> cache.CacheView[str, invites.InviteWithMetadata]:
        if not self._is_cache_enabled_for(config_api.CacheComponents.INVITES):
            return cache_utility.EmptyCacheView()

        return cache_utility.CacheMappingView(self._invite_entries.freeze(), builder=self._build_invite)

    @typing_extensions.override
    def get_invites_view_for_guild(
        self, guild: snowflakes.SnowflakeishOr[guilds.PartialGuild], /
    ) -> cache.CacheView[str, invites.InviteWithMetadata]:
        if not self._is_cache_enabled_for(config_api.CacheComponents.INVITES):
            return cache_utility.EmptyCacheView()

        guild_id = snowflakes.Snowflake(guild)
        guild_entry = self._guild_entries.get(guild_id)
        if not guild_entry or not guild_entry.invites:
            return cache_utility.EmptyCacheView()

        cached_invites = {code: self._invite_entries[code] for code in guild_entry.invites}
        return cache_utility.CacheMappingView(cached_invites, builder=self._build_invite)

    @typing_extensions.override
    def get_invites_view_for_channel(
        self,
        guild: snowflakes.SnowflakeishOr[guilds.PartialGuild],
        channel: snowflakes.SnowflakeishOr[channels_.PartialChannel],
        /,
    ) -> cache.CacheView[str, invites.InviteWithMetadata]:
        if not self._is_cache_enabled_for(config_api.CacheComponents.INVITES):
            return cache_utility.EmptyCacheView()

        guild_id = snowflakes.Snowflake(guild)
        channel_id = snowflakes.Snowflake(channel)
        guild_entry = self._guild_entries.get(guild_id)
        if not guild_entry or not guild_entry.invites:
            return cache_utility.EmptyCacheView()

        cached_invites = {
            invite.code: invite
            for invite in map(self._invite_entries.get, guild_entry.invites)
            if invite and invite.channel_id == channel_id
        }
        return cache_utility.CacheMappingView(cached_invites, builder=self._build_invite)

    @typing_extensions.override
    def set_invite(self, invite: invites.InviteWithMetadata, /) -> None:
        if not self._is_cache_enabled_for(config_api.CacheComponents.INVITES):
            return

        inviter: cache_utility.RefCell[users.User] | None = None
        if invite.inviter:
            inviter = self._set_user(invite.inviter)
            if invite.code not in self._invite_entries:
                self._increment_ref_count(inviter)

        target_user: cache_utility.RefCell[users.User] | None = None
        if invite.target_user:
            target_user = self._set_user(invite.target_user)
            if invite.code not in self._invite_entries:
                self._increment_ref_count(target_user)

        self._invite_entries[invite.code] = cache_utility.InviteData.build_from_entity(
            invite, inviter=inviter, target_user=target_user
        )
        if invite.guild_id:
            guild_entry = self._get_or_create_guild_record(invite.guild_id)

            if guild_entry.invites is None:
                guild_entry.invites = []

            guild_entry.invites.append(invite.code)

    @typing_extensions.override
    def update_invite(
        self, invite: invites.InviteWithMetadata, /
    ) -> tuple[invites.InviteWithMetadata | None, invites.InviteWithMetadata | None]:
        if not self._is_cache_enabled_for(config_api.CacheComponents.INVITES):
            return None, None

        cached_invite = self.get_invite(invite.code)
        self.set_invite(invite)
        return cached_invite, self.get_invite(invite.code)

    @typing_extensions.override
    def delete_me(self) -> users.OwnUser | None:
        cached_user = self._me
        self._me = None
        return cached_user

    @typing_extensions.override
    def get_me(self) -> users.OwnUser | None:
        return copy.copy(self._me)

    @typing_extensions.override
    def set_me(self, user: users.OwnUser, /) -> None:
        if self._is_cache_enabled_for(config_api.CacheComponents.ME):
            _LOGGER.debug("setting my user to %s", user)
            self._me = copy.copy(user)

    @typing_extensions.override
    def update_me(self, user: users.OwnUser, /) -> tuple[users.OwnUser | None, users.OwnUser | None]:
        if not self._is_cache_enabled_for(config_api.CacheComponents.ME):
            return None, None

        cached_user = self.get_me()
        self.set_me(user)
        return cached_user, self.get_me()

    def _build_member(self, member_data: cache_utility.RefCell[cache_utility.MemberData]) -> guilds.Member:
        return member_data.object.build_entity(self._app)

    @staticmethod
    def _can_remove_member(member: cache_utility.RefCell[cache_utility.MemberData]) -> bool:
        return member.ref_count < 1 and member.object.has_been_deleted

    def _garbage_collect_member(
        self,
        guild_record: cache_utility.GuildRecord,
        member: cache_utility.RefCell[cache_utility.MemberData],
        *,
        decrement: int | None = None,
        deleting: bool = False,
    ) -> cache_utility.RefCell[cache_utility.MemberData] | None:
        if deleting:
            member.object.has_been_deleted = True

        if decrement is not None:
            self._increment_ref_count(member, -decrement)

        user_id = member.object.user.object.id
        if not guild_record.members or user_id not in guild_record.members:
            return None

        if not self._can_remove_member(member):
            return None

        del guild_record.members[user_id]
        self._garbage_collect_user(member.object.user, decrement=1)

        if not guild_record.members:
            guild_record.members = None
            self._remove_guild_record_if_empty(member.object.guild_id, guild_record)

        return member

    @typing_extensions.override
    def clear_members(
        self,
    ) -> cache.CacheView[snowflakes.Snowflake, cache.CacheView[snowflakes.Snowflake, guilds.Member]]:
        if not self._is_cache_enabled_for(config_api.CacheComponents.MEMBERS):
            return cache_utility.EmptyCacheView()

        views = ((guild_id, self.clear_members_for_guild(guild_id)) for guild_id in self._guild_entries.freeze())
        return cache_utility.CacheMappingView({guild_id: view for guild_id, view in views if view})

    @typing_extensions.override
    def clear_members_for_guild(
        self, guild: snowflakes.SnowflakeishOr[guilds.PartialGuild], /
    ) -> cache.CacheView[snowflakes.Snowflake, guilds.Member]:
        if not self._is_cache_enabled_for(config_api.CacheComponents.MEMBERS):
            return cache_utility.EmptyCacheView()

        guild_id = snowflakes.Snowflake(guild)
        guild_record = self._guild_entries.get(guild_id)
        if not guild_record or not guild_record.members:
            return cache_utility.EmptyCacheView()

        cached_members = guild_record.members.freeze()
        members_gen = (self._garbage_collect_member(guild_record, m, deleting=True) for m in cached_members.values())
        # _garbage_collect_member will only return the member data object if they could be removed, else None.
        cached_members = {member.object.user.object.id: member for member in members_gen if member}
        self._remove_guild_record_if_empty(guild_id, guild_record)
        return cache_utility.CacheMappingView(cached_members, builder=self._build_member)  # type: ignore[type-var]

    @typing_extensions.override
    def delete_member(
        self,
        guild: snowflakes.SnowflakeishOr[guilds.PartialGuild],
        user: snowflakes.SnowflakeishOr[users.PartialUser],
        /,
    ) -> guilds.Member | None:
        if not self._is_cache_enabled_for(config_api.CacheComponents.MEMBERS):
            return None

        guild_id = snowflakes.Snowflake(guild)
        user_id = snowflakes.Snowflake(user)
        guild_record = self._guild_entries.get(guild_id)
        if not guild_record or not guild_record.members:
            return None

        member_data = guild_record.members.get(user_id)
        if not member_data:
            return None

        if not guild_record.members:
            guild_record.members = None
            self._remove_guild_record_if_empty(guild_id, guild_record)

        # _garbage_collect_member will only return the member data object if they could be removed, else None.
        garbage_collected = self._garbage_collect_member(guild_record, member_data, deleting=True)
        return self._build_member(member_data) if garbage_collected else None

    @typing_extensions.override
    def get_member(
        self,
        guild: snowflakes.SnowflakeishOr[guilds.PartialGuild],
        user: snowflakes.SnowflakeishOr[users.PartialUser],
        /,
    ) -> guilds.Member | None:
        if not self._is_cache_enabled_for(config_api.CacheComponents.MEMBERS):
            return None

        guild_id = snowflakes.Snowflake(guild)
        user_id = snowflakes.Snowflake(user)
        guild_record = self._guild_entries.get(guild_id)
        if not guild_record or not guild_record.members:
            return None

        member = guild_record.members.get(user_id)
        return self._build_member(member) if member else None

    @typing_extensions.override
    def get_members_view(
        self,
    ) -> cache.CacheView[snowflakes.Snowflake, cache.CacheView[snowflakes.Snowflake, guilds.Member]]:
        if not self._is_cache_enabled_for(config_api.CacheComponents.MEMBERS):
            return cache_utility.EmptyCacheView()

        views: typing.Mapping[snowflakes.Snowflake, cache.CacheView[snowflakes.Snowflake, guilds.Member]] = {
            guild_id: cache_utility.CacheMappingView(view.members.freeze(), builder=self._build_member)  # type: ignore[type-var]
            for guild_id, view in self._guild_entries.items()
            if view.members
        }
        return cache_utility.Cache3DMappingView(views)

    @typing_extensions.override
    def get_members_view_for_guild(
        self, guild_id: snowflakes.Snowflakeish, /
    ) -> cache.CacheView[snowflakes.Snowflake, guilds.Member]:
        if not self._is_cache_enabled_for(config_api.CacheComponents.MEMBERS):
            return cache_utility.EmptyCacheView()

        guild_id = snowflakes.Snowflake(guild_id)
        guild_record = self._guild_entries.get(guild_id)
        if not guild_record or not guild_record.members:
            return cache_utility.EmptyCacheView()

        cached_members = {
            user_id: member for user_id, member in guild_record.members.items() if not member.object.has_been_deleted
        }

        return cache_utility.CacheMappingView(cached_members, builder=self._build_member)  # type: ignore[type-var]

    @typing_extensions.override
    def set_member(self, member: guilds.Member, /) -> None:
        if not self._is_cache_enabled_for(config_api.CacheComponents.MEMBERS):
            return

        self._set_member(member, is_reference=False)

    def _set_member(
        self, member: guilds.Member, /, *, is_reference: bool = True
    ) -> cache_utility.RefCell[cache_utility.MemberData]:
        guild_record = self._get_or_create_guild_record(member.guild_id)
        user = self._set_user(member.user)
        member_data = cache_utility.MemberData.build_from_entity(member, user=user)

        if guild_record.members is None:  # TODO: test when this is not None
            guild_record.members = collections.FreezableDict()

        if member.user.id not in guild_record.members:
            self._increment_ref_count(member_data.user)

        try:
            member_data.has_been_deleted = False
            if is_reference:
                member_data.has_been_deleted = guild_record.members[member.id].object.has_been_deleted

            guild_record.members[member.id].object = member_data

        except KeyError:
            member_data.has_been_deleted = is_reference
            guild_record.members[member.id] = cache_utility.RefCell(member_data)

        return guild_record.members[member.id]

    @typing_extensions.override
    def update_member(self, member: guilds.Member, /) -> tuple[guilds.Member | None, guilds.Member | None]:
        if not self._is_cache_enabled_for(config_api.CacheComponents.MEMBERS):
            return None, None

        cached_member = self.get_member(member.guild_id, member.user.id)
        self.set_member(member)
        return cached_member, self.get_member(member.guild_id, member.user.id)

    def _build_presence(self, presence_data: cache_utility.MemberPresenceData) -> presences.MemberPresence:
        return presence_data.build_entity(self._app)

    def _garbage_collect_unknown_custom_emoji(
        self, emoji: cache_utility.RefCell[emojis.CustomEmoji], *, decrement: int | None = None
    ) -> None:
        if decrement is not None:
            self._increment_ref_count(emoji, -decrement)

        if emoji.ref_count < 1 and emoji.object.id in self._unknown_custom_emoji_entries:
            del self._unknown_custom_emoji_entries[emoji.object.id]

    def _remove_presence_assets(self, presence_data: cache_utility.MemberPresenceData) -> None:
        for activity_data in presence_data.activities:
            if isinstance(activity_data.emoji, cache_utility.RefCell):
                self._garbage_collect_unknown_custom_emoji(activity_data.emoji, decrement=1)

    @typing_extensions.override
    def clear_presences(
        self,
    ) -> cache.CacheView[snowflakes.Snowflake, cache.CacheView[snowflakes.Snowflake, presences.MemberPresence]]:
        if not self._is_cache_enabled_for(config_api.CacheComponents.PRESENCES):
            return cache_utility.EmptyCacheView()

        views = ((guild_id, self.clear_presences_for_guild(guild_id)) for guild_id in self._guild_entries.freeze())
        return cache_utility.CacheMappingView({guild_id: view for guild_id, view in views if view})

    @typing_extensions.override
    def clear_presences_for_guild(
        self, guild: snowflakes.SnowflakeishOr[guilds.PartialGuild], /
    ) -> cache.CacheView[snowflakes.Snowflake, presences.MemberPresence]:
        if not self._is_cache_enabled_for(config_api.CacheComponents.PRESENCES):
            return cache_utility.EmptyCacheView()

        guild_id = snowflakes.Snowflake(guild)
        guild_record = self._guild_entries.get(guild_id)
        if not guild_record or not guild_record.presences:
            return cache_utility.EmptyCacheView()

        cached_presences = guild_record.presences
        guild_record.presences = None

        for presence in cached_presences.values():
            self._remove_presence_assets(presence)

        self._remove_guild_record_if_empty(guild_id, guild_record)
        return cache_utility.CacheMappingView(cached_presences, builder=self._build_presence)

    @typing_extensions.override
    def delete_presence(
        self,
        guild: snowflakes.SnowflakeishOr[guilds.PartialGuild],
        user: snowflakes.SnowflakeishOr[users.PartialUser],
        /,
    ) -> presences.MemberPresence | None:
        if not self._is_cache_enabled_for(config_api.CacheComponents.PRESENCES):
            return None

        guild_id = snowflakes.Snowflake(guild)
        user_id = snowflakes.Snowflake(user)
        guild_record = self._guild_entries.get(guild_id)
        if not guild_record or not guild_record.presences:
            return None

        presence_data = guild_record.presences.pop(user_id, None)

        if not presence_data:
            return None

        self._remove_presence_assets(presence_data)

        if not guild_record.presences:
            guild_record.presences = None
            self._remove_guild_record_if_empty(guild_id, guild_record)

        return self._build_presence(presence_data)

    @typing_extensions.override
    def get_presence(
        self,
        guild: snowflakes.SnowflakeishOr[guilds.PartialGuild],
        user: snowflakes.SnowflakeishOr[users.PartialUser],
        /,
    ) -> presences.MemberPresence | None:
        if not self._is_cache_enabled_for(config_api.CacheComponents.PRESENCES):
            return None

        guild_id = snowflakes.Snowflake(guild)
        user_id = snowflakes.Snowflake(user)
        guild_record = self._guild_entries.get(guild_id)
        if not guild_record or not guild_record.presences:
            return None

        return self._build_presence(guild_record.presences[user_id]) if user_id in guild_record.presences else None

    @typing_extensions.override
    def get_presences_view(
        self,
    ) -> cache.CacheView[snowflakes.Snowflake, cache.CacheView[snowflakes.Snowflake, presences.MemberPresence]]:
        if not self._is_cache_enabled_for(config_api.CacheComponents.PRESENCES):
            return cache_utility.EmptyCacheView()

        views = {
            guild_id: cache_utility.CacheMappingView(guild_record.presences.freeze(), builder=self._build_presence)
            for guild_id, guild_record in self._guild_entries.items()
            if guild_record.presences
        }
        return cache_utility.Cache3DMappingView(views)

    @typing_extensions.override
    def get_presences_view_for_guild(
        self, guild: snowflakes.SnowflakeishOr[guilds.PartialGuild], /
    ) -> cache.CacheView[snowflakes.Snowflake, presences.MemberPresence]:
        if not self._is_cache_enabled_for(config_api.CacheComponents.PRESENCES):
            return cache_utility.EmptyCacheView()

        guild_record = self._guild_entries.get(snowflakes.Snowflake(guild))
        if not guild_record or not guild_record.presences:
            return cache_utility.EmptyCacheView()

        return cache_utility.CacheMappingView(guild_record.presences.freeze(), builder=self._build_presence)

    @typing_extensions.override
    def set_presence(self, presence: presences.MemberPresence, /) -> None:
        if not self._is_cache_enabled_for(config_api.CacheComponents.PRESENCES):
            return

        presence_data = cache_utility.MemberPresenceData.build_from_entity(presence)
        for activity, activity_data in zip(presence.activities, presence_data.activities):
            emoji = activity.emoji
            if not isinstance(emoji, emojis.CustomEmoji):
                continue

            if emoji.id in self._unknown_custom_emoji_entries:
                self._unknown_custom_emoji_entries[emoji.id].object = copy.copy(emoji)
                emoji_data = self._unknown_custom_emoji_entries[emoji.id]

            else:
                emoji_data = cache_utility.RefCell(copy.copy(emoji))
                self._unknown_custom_emoji_entries[emoji.id] = emoji_data

            self._increment_ref_count(emoji_data)
            activity_data.emoji = emoji_data

        guild_record = self._get_or_create_guild_record(presence.guild_id)
        if guild_record.presences is None:
            guild_record.presences = collections.FreezableDict()

        guild_record.presences[presence.user_id] = presence_data

    @typing_extensions.override
    def update_presence(
        self, presence: presences.MemberPresence, /
    ) -> tuple[presences.MemberPresence | None, presences.MemberPresence | None]:
        if not self._is_cache_enabled_for(config_api.CacheComponents.PRESENCES):
            return None, None

        cached_presence = self.get_presence(presence.guild_id, presence.user_id)
        self.set_presence(presence)
        return cached_presence, self.get_presence(presence.guild_id, presence.user_id)

    @typing_extensions.override
    def clear_roles(self) -> cache.CacheView[snowflakes.Snowflake, guilds.Role]:
        if not self._is_cache_enabled_for(config_api.CacheComponents.ROLES) or not self._role_entries:
            return cache_utility.EmptyCacheView()

        roles = self._role_entries
        self._role_entries = collections.FreezableDict()

        for guild_id, guild_record in self._guild_entries.freeze().items():
            if guild_record.roles:  # TODO: test coverage for when not this
                guild_record.roles = None
                self._remove_guild_record_if_empty(guild_id, guild_record)

        return cache_utility.CacheMappingView(roles)

    @typing_extensions.override
    def clear_roles_for_guild(
        self, guild: snowflakes.SnowflakeishOr[guilds.PartialGuild], /
    ) -> cache.CacheView[snowflakes.Snowflake, guilds.Role]:
        if not self._is_cache_enabled_for(config_api.CacheComponents.ROLES):
            return cache_utility.EmptyCacheView()

        guild_id = snowflakes.Snowflake(guild)
        guild_record = self._guild_entries.get(guild_id)
        if not guild_record or not guild_record.roles:
            return cache_utility.EmptyCacheView()

        view = cache_utility.CacheMappingView(
            {role_id: self._role_entries.pop(role_id) for role_id in guild_record.roles}
        )
        guild_record.roles = None
        self._remove_guild_record_if_empty(guild_id, guild_record)
        return view

    @typing_extensions.override
    def delete_role(self, role: snowflakes.SnowflakeishOr[guilds.PartialRole], /) -> guilds.Role | None:
        if not self._is_cache_enabled_for(config_api.CacheComponents.ROLES):
            return None

        role_id = snowflakes.Snowflake(role)
        role = self._role_entries.pop(role_id, None)
        if not role:
            return None

        guild_record = self._guild_entries.get(role.guild_id)
        if guild_record and guild_record.roles:
            guild_record.roles.remove(role_id)

            if not guild_record.roles:
                guild_record.roles = None
                self._remove_guild_record_if_empty(role.guild_id, guild_record)

        return role

    @typing_extensions.override
    def get_role(self, role: snowflakes.SnowflakeishOr[guilds.PartialRole], /) -> guilds.Role | None:
        if not self._is_cache_enabled_for(config_api.CacheComponents.ROLES):
            return None

        role = self._role_entries.get(snowflakes.Snowflake(role))
        return copy.copy(role) if role else None

    @typing_extensions.override
    def get_roles_view(self) -> cache.CacheView[snowflakes.Snowflake, guilds.Role]:
        if not self._is_cache_enabled_for(config_api.CacheComponents.ROLES):
            return cache_utility.EmptyCacheView()

        return cache_utility.CacheMappingView(self._role_entries.freeze())

    @typing_extensions.override
    def get_roles_view_for_guild(
        self, guild: snowflakes.SnowflakeishOr[guilds.PartialGuild], /
    ) -> cache.CacheView[snowflakes.Snowflake, guilds.Role]:
        if not self._is_cache_enabled_for(config_api.CacheComponents.ROLES):
            return cache_utility.EmptyCacheView()

        guild_record = self._guild_entries.get(snowflakes.Snowflake(guild))
        if not guild_record or not guild_record.roles:
            return cache_utility.EmptyCacheView()

        return cache_utility.CacheMappingView({role_id: self._role_entries[role_id] for role_id in guild_record.roles})

    @typing_extensions.override
    def set_role(self, role: guilds.Role, /) -> None:
        if not self._is_cache_enabled_for(config_api.CacheComponents.ROLES):
            return

        self._role_entries[role.id] = role
        guild_record = self._get_or_create_guild_record(role.guild_id)

        if guild_record.roles is None:  # TODO: test when this is not None
            guild_record.roles = collections.SnowflakeSet()

        guild_record.roles.add(role.id)

    @typing_extensions.override
    def update_role(self, role: guilds.Role, /) -> tuple[guilds.Role | None, guilds.Role | None]:
        if not self._is_cache_enabled_for(config_api.CacheComponents.ROLES):
            return None, None

        cached_role = self.get_role(role.id)
        self.set_role(role)
        return cached_role, self.get_role(role.id)

    @staticmethod
    def _can_remove_user(user_data: cache_utility.RefCell[users.User]) -> bool:
        return user_data.ref_count < 1

    def _garbage_collect_user(self, user: cache_utility.RefCell[users.User], *, decrement: int | None = None) -> None:
        if decrement is not None:
            self._increment_ref_count(user, -decrement)

        if self._can_remove_user(user) and user.object.id in self._user_entries:
            del self._user_entries[user.object.id]
            self._dm_channel_entries.pop(user.object.id, None)

    @typing_extensions.override
    def get_user(self, user: snowflakes.SnowflakeishOr[users.PartialUser], /) -> users.User | None:
        user = self._user_entries.get(snowflakes.Snowflake(user))
        return user.copy() if user else None

    @typing_extensions.override
    def get_users_view(self) -> cache.CacheView[snowflakes.Snowflake, users.User]:
        if not self._user_entries:
            return cache_utility.EmptyCacheView()

        cached_users = self._user_entries.freeze()
        unwrapper = typing.cast(
            "typing.Callable[[cache_utility.RefCell[users.User]], users.User]", cache_utility.unwrap_ref_cell
        )
        return cache_utility.CacheMappingView(cached_users, builder=unwrapper)  # type: ignore[type-var]

    def _set_user(self, user: users.User, /) -> cache_utility.RefCell[users.User]:
        try:
            self._user_entries[user.id].object = copy.copy(user)
            cell = self._user_entries[user.id]
        except KeyError:
            cell = cache_utility.RefCell(copy.copy(user))
            self._user_entries[user.id] = cell

        return cell

    def _build_voice_state(self, voice_data: cache_utility.VoiceStateData) -> voices.VoiceState:
        return voice_data.build_entity(self._app)

    @typing_extensions.override
    def clear_voice_states(
        self,
    ) -> cache.CacheView[snowflakes.Snowflake, cache.CacheView[snowflakes.Snowflake, voices.VoiceState]]:
        if not self._is_cache_enabled_for(config_api.CacheComponents.VOICE_STATES):
            return cache_utility.EmptyCacheView()

        views = ((guild_id, self.clear_voice_states_for_guild(guild_id)) for guild_id in self._guild_entries.freeze())
        return cache_utility.CacheMappingView({guild_id: view for guild_id, view in views if view})

    @typing_extensions.override
    def clear_voice_states_for_channel(
        self,
        guild: snowflakes.SnowflakeishOr[guilds.PartialGuild],
        channel: snowflakes.SnowflakeishOr[channels_.PartialChannel],
        /,
    ) -> cache.CacheView[snowflakes.Snowflake, voices.VoiceState]:
        if not self._is_cache_enabled_for(config_api.CacheComponents.VOICE_STATES):
            return cache_utility.EmptyCacheView()

        guild_id = snowflakes.Snowflake(guild)
        channel_id = snowflakes.Snowflake(channel)
        guild_record = self._guild_entries.get(guild_id)
        if not guild_record or not guild_record.voice_states:
            return cache_utility.EmptyCacheView()

        cached_voice_states = {}

        for user_id, voice_state in guild_record.voice_states.items():
            if voice_state.channel_id == channel_id:
                cached_voice_states[user_id] = voice_state
                if voice_state.member:
                    self._garbage_collect_member(guild_record, voice_state.member, decrement=1)

        if not guild_record.voice_states:
            guild_record.voice_states = None
            self._remove_guild_record_if_empty(guild_id, guild_record)

        return cache_utility.CacheMappingView(cached_voice_states, builder=self._build_voice_state)

    @typing_extensions.override
    def clear_voice_states_for_guild(
        self, guild: snowflakes.SnowflakeishOr[guilds.PartialGuild], /
    ) -> cache.CacheView[snowflakes.Snowflake, voices.VoiceState]:
        if not self._is_cache_enabled_for(config_api.CacheComponents.VOICE_STATES):
            return cache_utility.EmptyCacheView()

        guild_id = snowflakes.Snowflake(guild)
        guild_record = self._guild_entries.get(guild_id)

        if not guild_record or not guild_record.voice_states:
            return cache_utility.EmptyCacheView()

        cached_voice_states = guild_record.voice_states
        guild_record.voice_states = None

        for voice_state in cached_voice_states.values():
            if voice_state.member:
                self._garbage_collect_member(guild_record, voice_state.member, decrement=1)

        self._remove_guild_record_if_empty(guild_id, guild_record)
        return cache_utility.CacheMappingView(cached_voice_states, builder=self._build_voice_state)

    @typing_extensions.override
    def delete_voice_state(
        self,
        guild: snowflakes.SnowflakeishOr[guilds.PartialGuild],
        user: snowflakes.SnowflakeishOr[users.PartialUser],
        /,
    ) -> voices.VoiceState | None:
        if not self._is_cache_enabled_for(config_api.CacheComponents.VOICE_STATES):
            return None

        guild_id = snowflakes.Snowflake(guild)
        user_id = snowflakes.Snowflake(user)
        guild_record = self._guild_entries.get(guild_id)
        if not guild_record or not guild_record.voice_states:
            return None

        voice_state_data = guild_record.voice_states.pop(user_id, None)
        if not voice_state_data:
            return None

        if not guild_record.voice_states:
            guild_record.voice_states = None

        if voice_state_data.member:
            self._garbage_collect_member(guild_record, voice_state_data.member, decrement=1)

        self._remove_guild_record_if_empty(guild_id, guild_record)
        return self._build_voice_state(voice_state_data)

    @typing_extensions.override
    def get_voice_state(
        self,
        guild: snowflakes.SnowflakeishOr[guilds.PartialGuild],
        user: snowflakes.SnowflakeishOr[users.PartialUser],
        /,
    ) -> voices.VoiceState | None:
        if not self._is_cache_enabled_for(config_api.CacheComponents.VOICE_STATES):
            return None

        guild_id = snowflakes.Snowflake(guild)
        user_id = snowflakes.Snowflake(user)
        guild_record = self._guild_entries.get(guild_id)
        voice_data = guild_record.voice_states.get(user_id) if guild_record and guild_record.voice_states else None
        return self._build_voice_state(voice_data) if voice_data else None

    @typing_extensions.override
    def get_voice_states_view(
        self,
    ) -> cache.CacheView[snowflakes.Snowflake, cache.CacheView[snowflakes.Snowflake, voices.VoiceState]]:
        if not self._is_cache_enabled_for(config_api.CacheComponents.VOICE_STATES):
            return cache_utility.EmptyCacheView()

        views = {
            guild_id: cache_utility.CacheMappingView(
                guild_record.voice_states.freeze(), builder=self._build_voice_state
            )
            for guild_id, guild_record in self._guild_entries.items()
            if guild_record.voice_states
        }
        return cache_utility.Cache3DMappingView(views)

    @typing_extensions.override
    def get_voice_states_view_for_channel(
        self,
        guild: snowflakes.SnowflakeishOr[guilds.PartialGuild],
        channel: snowflakes.SnowflakeishOr[channels_.PartialChannel],
        /,
    ) -> cache.CacheView[snowflakes.Snowflake, voices.VoiceState]:
        if not self._is_cache_enabled_for(config_api.CacheComponents.VOICE_STATES):
            return cache_utility.EmptyCacheView()

        guild_id = snowflakes.Snowflake(guild)
        channel_id = snowflakes.Snowflake(channel)
        guild_record = self._guild_entries.get(guild_id)
        if not guild_record or not guild_record.voice_states:
            return cache_utility.EmptyCacheView()

        cached_voice_states = {
            user_id: voice_state
            for user_id, voice_state in guild_record.voice_states.items()
            if voice_state.channel_id == channel_id
        }

        return cache_utility.CacheMappingView(cached_voice_states, builder=self._build_voice_state)

    @typing_extensions.override
    def get_voice_states_view_for_guild(
        self, guild: snowflakes.SnowflakeishOr[guilds.PartialGuild], /
    ) -> cache.CacheView[snowflakes.Snowflake, voices.VoiceState]:
        if not self._is_cache_enabled_for(config_api.CacheComponents.VOICE_STATES):
            return cache_utility.EmptyCacheView()

        guild_record = self._guild_entries.get(snowflakes.Snowflake(guild))
        if not guild_record or not guild_record.voice_states:
            return cache_utility.EmptyCacheView()

        return cache_utility.CacheMappingView(guild_record.voice_states.freeze(), builder=self._build_voice_state)

    @typing_extensions.override
    def set_voice_state(self, voice_state: voices.VoiceState, /) -> None:
        if not self._is_cache_enabled_for(config_api.CacheComponents.VOICE_STATES):
            return

        guild_record = self._get_or_create_guild_record(voice_state.guild_id)

        if guild_record.voice_states is None:  # TODO: test when this is not None
            guild_record.voice_states = collections.FreezableDict()

        # If the member is missing for some reason, try our best to find
        # it and store it along side the voice state
        if voice_state.member:
            member = self._set_member(voice_state.member)
            if voice_state.user_id not in guild_record.voice_states:
                self._increment_ref_count(member)
        else:
            member = None

        voice_state_data = cache_utility.VoiceStateData.build_from_entity(voice_state, member=member)

        guild_record.voice_states[voice_state.user_id] = voice_state_data

    @typing_extensions.override
    def update_voice_state(
        self, voice_state: voices.VoiceState, /
    ) -> tuple[voices.VoiceState | None, voices.VoiceState | None]:
        if not self._is_cache_enabled_for(config_api.CacheComponents.VOICE_STATES):
            return None, None

        cached_voice_state = self.get_voice_state(voice_state.guild_id, voice_state.user_id)
        self.set_voice_state(voice_state)
        return cached_voice_state, self.get_voice_state(voice_state.guild_id, voice_state.user_id)

    def _build_message(self, message_data: cache_utility.RefCell[cache_utility.MessageData]) -> messages.Message:
        return message_data.object.build_entity(self._app)

    def _can_remove_message(self, message: cache_utility.RefCell[cache_utility.MessageData]) -> bool:
        return message.object.id not in self._message_entries and message.ref_count < 1

    def _garbage_collect_message(
        self,
        message: cache_utility.RefCell[cache_utility.MessageData],
        *,
        decrement: int | None = None,
        override_ref: bool = False,
    ) -> cache_utility.RefCell[cache_utility.MessageData] | None:
        if decrement is not None:
            self._increment_ref_count(message, -decrement)

        if not self._can_remove_message(message) or override_ref:
            return None

        self._garbage_collect_user(message.object.author, decrement=1)

        if message.object.member:
            guild_record = self._guild_entries.get(message.object.member.object.guild_id)
            if guild_record:
                self._garbage_collect_member(guild_record, message.object.member, decrement=1)

        if message.object.referenced_message:
            self._garbage_collect_message(message.object.referenced_message, decrement=1)

        if message.object.user_mentions:
            for user in message.object.user_mentions.values():
                self._garbage_collect_user(user, decrement=1)

        # If we got this far the message won't be in _message_entries as that'd infer that it hasn't been marked as
        # deleted yet.
        if message.object.id in self._referenced_messages:
            del self._referenced_messages[message.object.id]

        return message

    def _on_message_expire(self, message: cache_utility.RefCell[cache_utility.MessageData], /) -> None:
        if not self._garbage_collect_message(message):
            self._referenced_messages[message.object.id] = message

    @typing_extensions.override
    def clear_messages(self) -> cache.CacheView[snowflakes.Snowflake, messages.Message]:
        if not self._is_cache_enabled_for(config_api.CacheComponents.MESSAGES) or not self._message_entries:
            return cache_utility.EmptyCacheView()

        # As the only entry which references messages is other messages, this is enough for now.
        cached_messages = self._message_entries.freeze()
        self._message_entries.clear()
        cached_messages.update(self._referenced_messages)
        self._referenced_messages.clear()

        for message in cached_messages.values():
            self._garbage_collect_message(message, override_ref=True)

        return cache_utility.CacheMappingView(cached_messages, builder=self._build_message)  # type: ignore[type-var]

    @typing_extensions.override
    def delete_message(self, message: snowflakes.SnowflakeishOr[messages.PartialMessage], /) -> messages.Message | None:
        if not self._is_cache_enabled_for(config_api.CacheComponents.MESSAGES):
            return None

        message_id = snowflakes.Snowflake(message)
        message_data = self._message_entries.pop(message_id, None)

        if not message_data:
            return None

        if not self._garbage_collect_message(message_data):
            self._referenced_messages[message_id] = message_data
            return None

        return self._build_message(message_data)

    @typing_extensions.override
    def get_message(self, message: snowflakes.SnowflakeishOr[messages.PartialMessage], /) -> messages.Message | None:
        if not self._is_cache_enabled_for(config_api.CacheComponents.MESSAGES):
            return None

        message_id = snowflakes.Snowflake(message)
        message_data = self._message_entries.get(message_id) or self._referenced_messages.get(message_id)
        return self._build_message(message_data) if message_data else None

    @typing_extensions.override
    def get_messages_view(self) -> cache.CacheView[snowflakes.Snowflake, messages.Message]:
        if not self._is_cache_enabled_for(config_api.CacheComponents.MESSAGES):
            return cache_utility.EmptyCacheView()

        cached_messages = self._message_entries.freeze()
        cached_messages.update(self._referenced_messages)
        return cache_utility.CacheMappingView(cached_messages, builder=self._build_message)  # type: ignore[type-var]

    # We rather keep everything we can here inline.
    def _set_message(  # noqa: PLR0912
        self, message: messages.Message, /, *, is_reference: bool = True
    ) -> cache_utility.RefCell[cache_utility.MessageData]:
        author = self._set_user(message.author)
        member = self._set_member(message.member) if message.member else None

        user_mentions: undefined.UndefinedOr[
            typing.Mapping[snowflakes.Snowflake, cache_utility.RefCell[users.User]]
        ] = undefined.UNDEFINED
        if message.user_mentions is not undefined.UNDEFINED:
            user_mentions = {user_id: self._set_user(user) for user_id, user in message.user_mentions.items()}

        referenced_message: cache_utility.RefCell[cache_utility.MessageData] | None = None
        if message.referenced_message:
            reference_id = message.referenced_message.id
            referenced_message = self._message_entries.get(reference_id) or self._referenced_messages.get(reference_id)

            if referenced_message:
                # Since the message is partial, if we don't have it cached, there is nothing we can do about it
                referenced_message.object.update(message.referenced_message)

        # Only increment ref counts if this wasn't previously cached.
        if message.id not in self._referenced_messages and message.id not in self._message_entries:
            if member:
                self._increment_ref_count(member)

            if referenced_message:
                self._increment_ref_count(referenced_message)

            if user_mentions is not undefined.UNDEFINED:
                for user in user_mentions.values():
                    self._increment_ref_count(user)

        message_data = cache_utility.MessageData.build_from_entity(
            message, author=author, member=member, user_mentions=user_mentions, referenced_message=referenced_message
        )

        # Ensure any previously set message ref cell is in the right place before updating the cache.
        if not is_reference and message.id in self._referenced_messages:
            self._message_entries[message.id] = self._referenced_messages.pop(message.id)

        if message.id in self._message_entries:
            self._message_entries[message.id].object = message_data

        elif not is_reference:
            self._message_entries[message.id] = cache_utility.RefCell(message_data)

        elif message.id in self._referenced_messages:
            self._referenced_messages[message.id].object = message_data

        else:
            self._referenced_messages[message.id] = cache_utility.RefCell(message_data)

        return self._message_entries.get(message.id) or self._referenced_messages[message.id]

    @typing_extensions.override
    def set_message(self, message: messages.Message, /) -> None:
        if not self._is_cache_enabled_for(config_api.CacheComponents.MESSAGES):
            return

        self._set_message(message, is_reference=False)

    @typing_extensions.override
    def update_message(
        self, message: messages.PartialMessage | messages.Message, /
    ) -> tuple[messages.Message | None, messages.Message | None]:
        if not self._is_cache_enabled_for(config_api.CacheComponents.MESSAGES):
            return None, None

        cached_message = self.get_message(message.id)

        if isinstance(message, messages.Message):
            self.set_message(message)

        elif cached_message_data := self._message_entries.get(message.id) or self._referenced_messages.get(message.id):
            user_mentions: undefined.UndefinedOr[
                typing.Mapping[snowflakes.Snowflake, cache_utility.RefCell[users.User]]
            ] = undefined.UNDEFINED
            if message.user_mentions is not undefined.UNDEFINED:
                user_mentions = {user_id: self._set_user(user) for user_id, user in message.user_mentions.items()}

                # We want to ensure that any previously mentioned users are garbage collected if they're no longer
                # being mentioned.
                if cached_message_data.object.user_mentions is not undefined.UNDEFINED:
                    for user_id, user in cached_message_data.object.user_mentions.items():
                        if user_id not in user_mentions:
                            self._garbage_collect_user(user, decrement=1)

            cached_message_data.object.update(message, user_mentions=user_mentions)

        return cached_message, self.get_message(message.id)



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import mock
import pytest

from hikari import embeds


class TestEmbedResource:
    @pytest.fixture
    def resource(self):
        return embeds.EmbedResource(resource=mock.Mock())

    def test_url(self, resource):
        assert resource.url is resource.resource.url

    def test_filename(self, resource):
        assert resource.filename is resource.resource.filename

    def test_stream(self, resource):
        mock_executor = object()

        assert resource.stream(executor=mock_executor, head_only=True) is resource.resource.stream.return_value

        resource.resource.stream.assert_called_once_with(executor=mock_executor, head_only=True)


class TestEmbedResourceWithProxy:
    @pytest.fixture
    def resource_with_proxy(self):
        return embeds.EmbedResourceWithProxy(resource=mock.Mock(), proxy_resource=mock.Mock())

    def test_proxy_url(self, resource_with_proxy):
        assert resource_with_proxy.proxy_url is resource_with_proxy.proxy_resource.url

    def test_proxy_url_when_resource_is_none(self, resource_with_proxy):
        resource_with_proxy.proxy_resource = None
        assert resource_with_proxy.proxy_url is None

    def test_proxy_filename(self, resource_with_proxy):
        assert resource_with_proxy.proxy_filename is resource_with_proxy.proxy_resource.filename

    def test_proxy_filename_when_resource_is_none(self, resource_with_proxy):
        resource_with_proxy.proxy_resource = None
        assert resource_with_proxy.proxy_filename is None


class TestEmbed:
    def test_total_length_when_embed_is_empty(self):
        embed = embeds.Embed()
        assert embed.total_length() == 0

    def test_total_length_when_title_is_none(self):
        embed = embeds.Embed(title=None)
        assert embed.total_length() == 0

    def test_total_length_title(self):
        embed = embeds.Embed(title="title")
        assert embed.total_length() == 5

    def test_total_length_when_description_is_none(self):
        embed = embeds.Embed(description=None)
        assert embed.total_length() == 0

    def test_total_length_description(self):
        embed = embeds.Embed(description="description")
        assert embed.total_length() == 11

    def test_total_length_author_name(self):
        embed = embeds.Embed().set_author(name="author name")
        assert embed.total_length() == 11

    def test_total_length_footer_text(self):
        embed = embeds.Embed().set_footer(text="footer text")
        assert embed.total_length() == 11

    def test_total_length_field_name(self):
        embed = embeds.Embed().add_field(name="field name", value="")
        assert embed.total_length() == 10

    def test_total_length_field_value(self):
        embed = embeds.Embed().add_field(name="", value="field value")
        assert embed.total_length() == 11

    def test_total_length_all(self):
        embed = embeds.Embed(title="title", description="description")
        embed.set_author(name="author name")
        embed.set_footer(text="footer text")
        embed.add_field(name="field name 1", value="field value 2")
        embed.add_field(name="field name 3", value="field value 4")
        assert embed.total_length() == 88

    def test_adding_field_with_none_values(self):
        embed = embeds.Embed().add_field(name=None, value=None)
        assert embed.total_length() == 0
        assert embed.fields[0].name == ""
        assert embed.fields[0].value == ""



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import datetime

import mock
import pytest

from hikari import channels as channels_
from hikari import colors
from hikari import guilds
from hikari import permissions
from hikari import snowflakes
from hikari import undefined
from hikari import urls
from hikari import users
from hikari.impl import gateway_bot
from hikari.internal import routes
from hikari.internal import time
from tests.hikari import hikari_test_helpers


@pytest.fixture
def mock_app():
    return mock.Mock(spec_set=gateway_bot.GatewayBot)


class TestPartialRole:
    @pytest.fixture
    def model(self, mock_app):
        return guilds.PartialRole(app=mock_app, id=snowflakes.Snowflake(1106913972), name="The Big Cool")

    def test_str_operator(self, model):
        assert str(model) == "The Big Cool"

    def test_mention_property(self, model):
        assert model.mention == "<@&1106913972>"


def test_PartialApplication_str_operator():
    mock_application = mock.Mock(guilds.PartialApplication)
    mock_application.name = "beans"
    assert guilds.PartialApplication.__str__(mock_application) == "beans"


class TestPartialApplication:
    @pytest.fixture
    def model(self):
        return hikari_test_helpers.mock_class_namespace(
            guilds.PartialApplication, init_=False, slots_=False, id=123, icon_hash="ahashicon"
        )()

    def test_make_icon_url_format_set_to_deprecated_ext_argument_if_provided(self, model):
        with mock.patch.object(
            routes, "CDN_APPLICATION_ICON", new=mock.Mock(compile_to_file=mock.Mock(return_value="file"))
        ) as route:
            assert model.make_icon_url(ext="JPEG") == "file"

        route.compile_to_file.assert_called_once_with(
            urls.CDN_URL, application_id=123, hash="ahashicon", size=4096, file_format="JPEG", lossless=True
        )

    def test_icon_url_property(self, model):
        model.make_icon_url = mock.Mock(return_value="url")

        assert model.icon_url == "url"

        model.make_icon_url.assert_called_once_with()

    def test_make_icon_url_when_hash_is_None(self, model):
        model.icon_hash = None

        with mock.patch.object(
            routes, "CDN_APPLICATION_ICON", new=mock.Mock(compile_to_file=mock.Mock(return_value="file"))
        ) as route:
            assert model.make_icon_url(file_format="JPEG", size=1) is None

        route.compile_to_file.assert_not_called()

    def test_make_icon_url_when_hash_is_not_None(self, model):
        with mock.patch.object(
            routes, "CDN_APPLICATION_ICON", new=mock.Mock(compile_to_file=mock.Mock(return_value="file"))
        ) as route:
            assert model.make_icon_url(file_format="JPEG", size=1) == "file"

        route.compile_to_file.assert_called_once_with(
            urls.CDN_URL, application_id=123, hash="ahashicon", size=1, file_format="JPEG", lossless=True
        )


class TestIntegrationAccount:
    @pytest.fixture
    def model(self, mock_app):
        return guilds.IntegrationAccount(id="foo", name="bar")

    def test_str_operator(self, model):
        assert str(model) == "bar"


class TestPartialIntegration:
    @pytest.fixture
    def model(self, mock_app):
        return guilds.PartialIntegration(
            account=mock.Mock(return_value=guilds.IntegrationAccount),
            id=snowflakes.Snowflake(69420),
            name="nice",
            type="illegal",
        )

    def test_str_operator(self, model):
        assert str(model) == "nice"


class TestRole:
    @pytest.fixture
    def model(self, mock_app):
        return guilds.Role(
            app=mock_app,
            id=snowflakes.Snowflake(979899100),
            name="@everyone",
            color=colors.Color(0x1A2B3C),
            guild_id=snowflakes.Snowflake(112233),
            is_hoisted=False,
            icon_hash="icon_hash",
            unicode_emoji=None,
            is_managed=False,
            is_mentionable=True,
            permissions=permissions.Permissions.CONNECT,
            position=12,
            bot_id=None,
            integration_id=None,
            is_premium_subscriber_role=False,
            is_guild_linked_role=True,
            subscription_listing_id=snowflakes.Snowflake(10),
            is_available_for_purchase=True,
        )

    def test_colour_property(self, model):
        assert model.colour == colors.Color(0x1A2B3C)

    def test_make_icon_url_format_set_to_deprecated_ext_argument_if_provided(self, model):
        with mock.patch.object(
            routes, "CDN_ROLE_ICON", new=mock.Mock(compile_to_file=mock.Mock(return_value="file"))
        ) as route:
            assert model.make_icon_url(ext="JPEG") == "file"

        route.compile_to_file.assert_called_once_with(
            urls.CDN_URL, role_id=979899100, hash="icon_hash", size=4096, file_format="JPEG", lossless=True
        )

    def test_icon_url_property(self, model):
        with mock.patch.object(guilds.Role, "make_icon_url") as make_icon_url:
            assert model.icon_url == make_icon_url.return_value

            model.make_icon_url.assert_called_once_with()

    def test_mention_property(self, model):
        assert model.mention == "<@&979899100>"

    def test_mention_property_when_is_everyone_role(self, model):
        model.id = model.guild_id
        assert model.mention == "@everyone"

    def test_make_icon_url_when_hash_is_None(self, model):
        model.icon_hash = None

        with mock.patch.object(
            routes, "CDN_ROLE_ICON", new=mock.Mock(compile_to_file=mock.Mock(return_value="file"))
        ) as route:
            assert model.make_icon_url(file_format="JPEG", size=1) is None

        route.compile_to_file.assert_not_called()

    def test_make_icon_url_when_hash_is_not_None(self, model):
        with mock.patch.object(
            routes, "CDN_ROLE_ICON", new=mock.Mock(compile_to_file=mock.Mock(return_value="file"))
        ) as route:
            assert model.make_icon_url(file_format="JPEG", size=1) == "file"

        route.compile_to_file.assert_called_once_with(
            urls.CDN_URL, role_id=979899100, hash="icon_hash", size=1, file_format="JPEG", lossless=True
        )


class TestGuildWidget:
    @pytest.fixture
    def model(self, mock_app):
        return guilds.GuildWidget(app=mock_app, channel_id=snowflakes.Snowflake(420), is_enabled=True)

    def test_app_property(self, model, mock_app):
        assert model.app is mock_app

    def test_channel_property(self, model):
        assert model.channel_id == snowflakes.Snowflake(420)

    def test_is_enabled_property(self, model):
        assert model.is_enabled is True

    @pytest.mark.asyncio
    async def test_fetch_channel(self, model):
        mock_channel = mock.Mock(channels_.GuildChannel)
        model.app.rest.fetch_channel = mock.AsyncMock(return_value=mock_channel)

        assert await model.fetch_channel() is model.app.rest.fetch_channel.return_value
        model.app.rest.fetch_channel.assert_awaited_once_with(420)

    @pytest.mark.asyncio
    async def test_fetch_channel_when_None(self, model):
        model.app.rest.fetch_channel = mock.AsyncMock()
        model.channel_id = None

        assert await model.fetch_channel() is None


class TestMember:
    @pytest.fixture
    def mock_user(self):
        return mock.Mock(id=snowflakes.Snowflake(123))

    @pytest.fixture
    def model(self, mock_user):
        return guilds.Member(
            guild_id=snowflakes.Snowflake(456),
            is_deaf=True,
            is_mute=True,
            is_pending=False,
            joined_at=datetime.datetime.now().astimezone(),
            nickname="davb",
            guild_avatar_decoration=users.AvatarDecoration(
                asset_hash="dimmadab",
                sku_id=snowflakes.Snowflake(123456789),
                expires_at=datetime.datetime(2025, 4, 4, 1, 1, 1),
            ),
            guild_avatar_hash="dab",
            guild_banner_hash="dimma",
            premium_since=None,
            role_ids=[snowflakes.Snowflake(456), snowflakes.Snowflake(1234)],
            user=mock_user,
            raw_communication_disabled_until=None,
            guild_flags=guilds.GuildMemberFlags.NONE,
        )

    def test_str_operator(self, model, mock_user):
        assert str(model) == str(mock_user)

    def test_app_property(self, model):
        assert model.app is model.user.app

    def test_id_property(self, model):
        assert model.id is model.user.id

    def test_username_property(self, model):
        assert model.username is model.user.username

    def test_discriminator_property(self, model):
        assert model.discriminator is model.user.discriminator

    def test_avatar_hash_property(self, model):
        assert model.avatar_hash is model.user.avatar_hash

    def test_is_bot_property(self, model):
        assert model.is_bot is model.user.is_bot

    def test_is_system_property(self, model):
        assert model.is_system is model.user.is_system

    def test_flags_property(self, model):
        assert model.flags is model.user.flags

    def test_display_avatar_decoration_property_when_guild_avatar_decoration_is_set(self, model):
        assert model.display_avatar_decoration is model.guild_avatar_decoration

    def test_display_avatar_decoration_property_when_guild_avatar_decoration_is_None(self, model):
        model.guild_avatar_decoration = None
        assert model.display_avatar_decoration is model.user.display_avatar_decoration

    def test_avatar_url_property(self, model):
        assert model.avatar_url is model.user.make_avatar_url()

    def test_display_avatar_url_when_guild_hash_is_None(self, model):
        with mock.patch.object(guilds.Member, "make_guild_avatar_url") as mock_make_guild_avatar_url:
            assert model.display_avatar_url is mock_make_guild_avatar_url.return_value

    def test_display_guild_avatar_url_when_guild_hash_is_not_None(self, model):
        with mock.patch.object(guilds.Member, "make_guild_avatar_url", return_value=None):
            assert model.display_avatar_url is model.user.display_avatar_url

    def test_display_banner_url_when_guild_hash_is_None(self, model):
        with mock.patch.object(guilds.Member, "make_guild_banner_url") as mock_make_guild_banner_url:
            assert model.display_banner_url is mock_make_guild_banner_url.return_value

    def test_display_banner_url_when_guild_hash_is_not_None(self, model):
        with mock.patch.object(guilds.Member, "make_guild_banner_url", return_value=None):
            assert model.display_banner_url is model.user.display_banner_url

    def test_banner_hash_property(self, model):
        assert model.banner_hash is model.user.banner_hash

    def test_banner_url_property(self, model):
        assert model.banner_url is model.user.make_banner_url()

    def test_accent_color_property(self, model):
        assert model.accent_color is model.user.accent_color

    def test_guild_avatar_url_property(self, model):
        with mock.patch.object(guilds.Member, "make_guild_avatar_url") as make_guild_avatar_url:
            assert model.guild_avatar_url is make_guild_avatar_url.return_value

    def test_guild_banner_url_property(self, model):
        with mock.patch.object(guilds.Member, "make_guild_banner_url") as make_guild_banner_url:
            assert model.guild_banner_url is make_guild_banner_url.return_value

    def test_communication_disabled_until(self, model):
        model.raw_communication_disabled_until = datetime.datetime(2021, 11, 22)

        with mock.patch.object(time, "utc_datetime", return_value=datetime.datetime(2021, 10, 18)):
            assert model.communication_disabled_until() == datetime.datetime(2021, 11, 22)

    def test_communication_disabled_until_when_raw_communication_disabled_until_is_None(self, model):
        model.raw_communication_disabled_until = None

        with mock.patch.object(time, "utc_datetime", return_value=datetime.datetime(2021, 10, 18)):
            assert model.communication_disabled_until() is None

    def test_communication_disabled_until_when_raw_communication_disabled_until_is_in_the_past(self, model):
        model.raw_communication_disabled_until = datetime.datetime(2021, 10, 18)

        with mock.patch.object(time, "utc_datetime", return_value=datetime.datetime(2021, 11, 22)):
            assert model.communication_disabled_until() is None

    def test_make_avatar_url_format_set_to_deprecated_ext_argument_if_provided(self, model):
        with mock.patch.object(
            routes, "CDN_MEMBER_AVATAR", new=mock.Mock(compile_to_file=mock.Mock(return_value="file"))
        ) as route:
            assert model.make_guild_avatar_url(ext="JPEG") == "file"

        route.compile_to_file.assert_called_once_with(
            urls.CDN_URL, guild_id=456, user_id=123, hash="dab", size=4096, file_format="JPEG", lossless=True
        )

    def test_make_avatar_url(self, model):
        result = model.make_avatar_url(file_format="PNG", size=4096)

        model.user.make_avatar_url.assert_called_once_with(
            file_format="PNG", size=4096, lossless=True, ext=undefined.UNDEFINED
        )
        assert result is model.user.make_avatar_url.return_value

    def test_make_guild_avatar_url_when_no_hash(self, model):
        model.guild_avatar_hash = None
        assert model.make_guild_avatar_url(file_format="PNG", size=1024) is None

    def test_make_guild_avatar_url_when_format_is_None_and_avatar_hash_is_for_animated(self, model):
        model.guild_avatar_hash = "a_18dnf8dfbakfdh"

        with mock.patch.object(
            routes, "CDN_MEMBER_AVATAR", new=mock.Mock(compile_to_file=mock.Mock(return_value="file"))
        ) as route:
            assert model.make_guild_avatar_url(file_format=None, size=4096) == "file"

        route.compile_to_file.assert_called_once_with(
            urls.CDN_URL,
            guild_id=model.guild_id,
            user_id=model.id,
            hash=model.guild_avatar_hash,
            size=4096,
            file_format="GIF",
            lossless=True,
        )

    def test_make_guild_avatar_url_when_format_is_None_and_avatar_hash_is_not_for_animated(self, model):
        model.guild_avatar_hash = "18dnf8dfbakfdh"

        with mock.patch.object(
            routes, "CDN_MEMBER_AVATAR", new=mock.Mock(compile_to_file=mock.Mock(return_value="file"))
        ) as route:
            assert model.make_guild_avatar_url(file_format=None, size=4096) == "file"

        route.compile_to_file.assert_called_once_with(
            urls.CDN_URL,
            guild_id=model.guild_id,
            user_id=model.id,
            hash=model.guild_avatar_hash,
            size=4096,
            file_format="PNG",
            lossless=True,
        )

    def test_make_guild_avatar_url_with_all_args(self, model):
        model.guild_avatar_hash = "18dnf8dfbakfdh"

        with mock.patch.object(
            routes, "CDN_MEMBER_AVATAR", new=mock.Mock(compile_to_file=mock.Mock(return_value="file"))
        ) as route:
            assert model.make_guild_avatar_url(file_format="URL", size=4096) == "file"

        route.compile_to_file.assert_called_once_with(
            urls.CDN_URL,
            guild_id=model.guild_id,
            user_id=model.id,
            hash=model.guild_avatar_hash,
            size=4096,
            file_format="URL",
            lossless=True,
        )

    def test_make_banner_url_format_set_to_deprecated_ext_argument_if_provided(self, model):
        with mock.patch.object(
            routes, "CDN_MEMBER_BANNER", new=mock.Mock(compile_to_file=mock.Mock(return_value="file"))
        ) as route:
            assert model.make_guild_banner_url(ext="JPEG") == "file"

        route.compile_to_file.assert_called_once_with(
            urls.CDN_URL,
            user_id=model.id,
            guild_id=model.guild_id,
            hash=model.guild_banner_hash,
            size=4096,
            file_format="JPEG",
            lossless=True,
        )

    def test_make_banner_url(self, model):
        result = model.make_banner_url(file_format="PNG", size=4096)

        model.user.make_banner_url.assert_called_once_with(
            file_format="PNG", size=4096, lossless=True, ext=undefined.UNDEFINED
        )
        assert result is model.user.make_banner_url.return_value

    def test_make_guild_banner_url_when_no_hash(self, model):
        model.guild_banner_hash = None
        assert model.make_guild_banner_url(file_format="PNG", size=1024) is None

    def test_make_guild_banner_url_when_format_is_None_and_banner_hash_is_for_animated(self, model):
        model.guild_banner_hash = "a_18dnf8dfbakfdh"

        with mock.patch.object(
            routes, "CDN_MEMBER_BANNER", new=mock.Mock(compile_to_file=mock.Mock(return_value="file"))
        ) as route:
            assert model.make_guild_banner_url(file_format=None, size=4096) == "file"

        route.compile_to_file.assert_called_once_with(
            urls.CDN_URL,
            user_id=model.id,
            guild_id=model.guild_id,
            hash=model.guild_banner_hash,
            size=4096,
            file_format="GIF",
            lossless=True,
        )

    def test_make_guild_banner_url_when_format_is_None_and_banner_hash_is_not_for_animated(self, model):
        model.guild_banner_hash = "18dnf8dfbakfdh"

        with mock.patch.object(
            routes, "CDN_MEMBER_BANNER", new=mock.Mock(compile_to_file=mock.Mock(return_value="file"))
        ) as route:
            assert model.make_guild_banner_url(file_format=None, size=4096) == "file"

        route.compile_to_file.assert_called_once_with(
            urls.CDN_URL,
            user_id=model.id,
            guild_id=model.guild_id,
            hash=model.guild_banner_hash,
            size=4096,
            file_format="PNG",
            lossless=True,
        )

    def test_make_guild_banner_url_with_all_args(self, model):
        model.guild_banner_hash = "18dnf8dfbakfdh"

        with mock.patch.object(
            routes, "CDN_MEMBER_BANNER", new=mock.Mock(compile_to_file=mock.Mock(return_value="file"))
        ) as route:
            assert model.make_guild_banner_url(file_format="URL", size=4096) == "file"

        route.compile_to_file.assert_called_once_with(
            urls.CDN_URL,
            guild_id=model.guild_id,
            user_id=model.id,
            hash=model.guild_banner_hash,
            size=4096,
            file_format="URL",
            lossless=True,
        )

    @pytest.mark.asyncio
    async def test_fetch_dm_channel(self, model):
        model.user.fetch_dm_channel = mock.AsyncMock()

        assert await model.fetch_dm_channel() is model.user.fetch_dm_channel.return_value

        model.user.fetch_dm_channel.assert_awaited_once_with()

    @pytest.mark.asyncio
    async def test_fetch_self(self, model):
        model.user.app.rest.fetch_member = mock.AsyncMock()

        assert await model.fetch_self() is model.user.app.rest.fetch_member.return_value

        model.user.app.rest.fetch_member.assert_awaited_once_with(456, 123)

    @pytest.mark.asyncio
    async def test_fetch_roles(self, model):
        model.user.app.rest.fetch_roles = mock.AsyncMock()
        await model.fetch_roles()
        model.user.app.rest.fetch_roles.assert_awaited_once_with(456)

    @pytest.mark.asyncio
    async def test_ban(self, model):
        model.app.rest.ban_user = mock.AsyncMock()

        await model.ban(delete_message_seconds=600, reason="bored")

        model.app.rest.ban_user.assert_awaited_once_with(456, 123, delete_message_seconds=600, reason="bored")

    @pytest.mark.asyncio
    async def test_unban(self, model):
        model.app.rest.unban_user = mock.AsyncMock()

        await model.unban(reason="Unbored")

        model.app.rest.unban_user.assert_awaited_once_with(456, 123, reason="Unbored")

    @pytest.mark.asyncio
    async def test_kick(self, model):
        model.app.rest.kick_user = mock.AsyncMock()

        await model.kick(reason="bored")

        model.app.rest.kick_user.assert_awaited_once_with(456, 123, reason="bored")

    @pytest.mark.asyncio
    async def test_add_role(self, model):
        model.app.rest.add_role_to_member = mock.AsyncMock()

        await model.add_role(563412, reason="Promoted")

        model.app.rest.add_role_to_member.assert_awaited_once_with(456, 123, 563412, reason="Promoted")

    @pytest.mark.asyncio
    async def test_remove_role(self, model):
        model.app.rest.remove_role_from_member = mock.AsyncMock()

        await model.remove_role(563412, reason="Demoted")

        model.app.rest.remove_role_from_member.assert_awaited_once_with(456, 123, 563412, reason="Demoted")

    @pytest.mark.asyncio
    async def test_edit(self, model):
        model.app.rest.edit_member = mock.AsyncMock()
        disabled_until = datetime.datetime(2021, 11, 17)
        edit = await model.edit(
            nickname="Imposter",
            roles=[123, 432, 345],
            mute=False,
            deaf=True,
            voice_channel=4321245,
            communication_disabled_until=disabled_until,
            reason="I'm God",
        )

        model.app.rest.edit_member.assert_awaited_once_with(
            456,
            123,
            nickname="Imposter",
            roles=[123, 432, 345],
            mute=False,
            deaf=True,
            voice_channel=4321245,
            communication_disabled_until=disabled_until,
            reason="I'm God",
        )

        assert edit == model.app.rest.edit_member.return_value

    def test_default_avatar_url_property(self, model):
        assert model.default_avatar_url is model.user.default_avatar_url

    def test_display_name_property_when_nickname(self, model):
        assert model.display_name == "davb"

    def test_display_name_property_when_no_nickname(self, model):
        model.nickname = None
        assert model.display_name is model.user.global_name

    def test_mention_property(self, model):
        assert model.mention == model.user.mention

    def test_get_guild(self, model):
        guild = mock.Mock(id=456)
        model.user.app.cache.get_guild.side_effect = [guild]

        assert model.get_guild() == guild

        model.user.app.cache.get_guild.assert_has_calls([mock.call(456)])

    def test_get_guild_when_guild_not_in_cache(self, model):
        model.user.app.cache.get_guild.side_effect = [None]

        assert model.get_guild() is None

        model.user.app.cache.get_guild.assert_has_calls([mock.call(456)])

    def test_get_guild_when_no_cache_trait(self, model):
        model.user.app = object()

        assert model.get_guild() is None

    def test_get_roles(self, model):
        role1 = mock.Mock(id=321, position=2)
        role2 = mock.Mock(id=654, position=1)
        model.user.app.cache.get_role.side_effect = [role1, role2]
        model.role_ids = [321, 654]

        assert model.get_roles() == [role1, role2]

        model.user.app.cache.get_role.assert_has_calls([mock.call(321), mock.call(654)])

    def test_get_roles_when_role_ids_not_in_cache(self, model):
        role = mock.Mock(id=456, position=1)
        model.user.app.cache.get_role.side_effect = [None, role]
        model.role_ids = [321, 456]

        assert model.get_roles() == [role]

        model.user.app.cache.get_role.assert_has_calls([mock.call(321), mock.call(456)])

    def test_get_roles_when_empty_cache(self, model):
        model.role_ids = [132, 432]
        model.user.app.cache.get_role.side_effect = [None, None]

        assert model.get_roles() == []

        model.user.app.cache.get_role.assert_has_calls([mock.call(132), mock.call(432)])

    def test_get_roles_when_no_cache_trait(self, model):
        model.user.app = object()

        assert model.get_roles() == []

    def test_get_top_role(self, model):
        role1 = mock.Mock(id=321, position=2)
        role2 = mock.Mock(id=654, position=1)

        with mock.patch.object(guilds.Member, "get_roles", return_value=[role1, role2]):
            assert model.get_top_role() is role1

    def test_get_top_role_when_roles_is_empty(self, model):
        with mock.patch.object(guilds.Member, "get_roles", return_value=[]):
            assert model.get_top_role() is None

    def test_get_presence(self, model):
        assert model.get_presence() is model.user.app.cache.get_presence.return_value
        model.user.app.cache.get_presence.assert_called_once_with(456, 123)

    def test_get_presence_when_no_cache_trait(self, model):
        model.user.app = object()
        assert model.get_presence() is None


class TestPartialGuild:
    @pytest.fixture
    def model(self, mock_app):
        return guilds.PartialGuild(app=mock_app, id=snowflakes.Snowflake(90210), icon_hash="yeet", name="hikari")

    def test_str_operator(self, model):
        assert str(model) == "hikari"

    def test_shard_id_property(self, model):
        model.app.shard_count = 4
        assert model.shard_id == 0

    def test_shard_id_when_not_shard_aware(self, model):
        model.app = object()

        assert model.shard_id is None

    def test_make_icon_url_format_set_to_deprecated_ext_argument_if_provided(self, model):
        with mock.patch.object(
            routes, "CDN_GUILD_ICON", new=mock.Mock(compile_to_file=mock.Mock(return_value="file"))
        ) as route:
            assert model.make_icon_url(ext="JPEG") == "file"

        route.compile_to_file.assert_called_once_with(
            urls.CDN_URL, guild_id=90210, hash="yeet", size=4096, file_format="JPEG", lossless=True
        )

    def test_icon_url(self, model):
        icon = object()

        with mock.patch.object(guilds.PartialGuild, "make_icon_url", return_value=icon):
            assert model.icon_url is icon

    def test_make_icon_url_when_no_hash(self, model):
        model.icon_hash = None

        assert model.make_icon_url(file_format="PNG", size=2048) is None

    def test_make_icon_url_when_format_is_None_and_avatar_hash_is_for_animated(self, model):
        model.icon_hash = "a_yeet"

        with mock.patch.object(
            routes, "CDN_GUILD_ICON", new=mock.Mock(compile_to_file=mock.Mock(return_value="file"))
        ) as route:
            assert model.make_icon_url(file_format=None, size=1024) == "file"

        route.compile_to_file.assert_called_once_with(
            urls.CDN_URL, guild_id=90210, hash="a_yeet", size=1024, file_format="GIF", lossless=True
        )

    def test_make_icon_url_when_format_is_None_and_avatar_hash_is_not_for_animated(self, model):
        with mock.patch.object(
            routes, "CDN_GUILD_ICON", new=mock.Mock(compile_to_file=mock.Mock(return_value="file"))
        ) as route:
            assert model.make_icon_url(file_format=None, size=4096) == "file"

        route.compile_to_file.assert_called_once_with(
            urls.CDN_URL, guild_id=90210, hash="yeet", size=4096, file_format="PNG", lossless=True
        )

    def test_make_icon_url_with_all_args(self, model):
        with mock.patch.object(
            routes, "CDN_GUILD_ICON", new=mock.Mock(compile_to_file=mock.Mock(return_value="file"))
        ) as route:
            assert model.make_icon_url(file_format="URL", size=2048) == "file"

        route.compile_to_file.assert_called_once_with(
            urls.CDN_URL, guild_id=90210, hash="yeet", size=2048, file_format="URL", lossless=True
        )

    @pytest.mark.asyncio
    async def test_kick(self, model):
        model.app.rest.kick_user = mock.AsyncMock()
        await model.kick(4321, reason="Go away!")

        model.app.rest.kick_user.assert_awaited_once_with(90210, 4321, reason="Go away!")

    @pytest.mark.asyncio
    async def test_ban(self, model):
        model.app.rest.ban_user = mock.AsyncMock()

        await model.ban(4321, delete_message_seconds=864000, reason="Go away!")

        model.app.rest.ban_user.assert_awaited_once_with(90210, 4321, delete_message_seconds=864000, reason="Go away!")

    @pytest.mark.asyncio
    async def test_unban(self, model):
        model.app.rest.unban_user = mock.AsyncMock()
        await model.unban(4321, reason="Comeback!!")

        model.app.rest.unban_user.assert_awaited_once_with(90210, 4321, reason="Comeback!!")

    @pytest.mark.asyncio
    async def test_edit(self, model):
        model.app.rest.edit_guild = mock.AsyncMock()
        edited_guild = await model.edit(
            name="chad server",
            verification_level=guilds.GuildVerificationLevel.LOW,
            default_message_notifications=guilds.GuildMessageNotificationsLevel.ALL_MESSAGES,
            explicit_content_filter_level=guilds.GuildExplicitContentFilterLevel.DISABLED,
            owner=6996,
            afk_timeout=400,
            preferred_locale="us-en",
            features=[guilds.GuildFeature.COMMUNITY, guilds.GuildFeature.RAID_ALERTS_DISABLED],
            reason="beep boop",
        )

        model.app.rest.edit_guild.assert_awaited_once_with(
            90210,
            name="chad server",
            verification_level=guilds.GuildVerificationLevel.LOW,
            default_message_notifications=guilds.GuildMessageNotificationsLevel.ALL_MESSAGES,
            explicit_content_filter_level=guilds.GuildExplicitContentFilterLevel.DISABLED,
            afk_channel=undefined.UNDEFINED,
            afk_timeout=400,
            icon=undefined.UNDEFINED,
            owner=6996,
            splash=undefined.UNDEFINED,
            banner=undefined.UNDEFINED,
            system_channel=undefined.UNDEFINED,
            rules_channel=undefined.UNDEFINED,
            public_updates_channel=undefined.UNDEFINED,
            preferred_locale="us-en",
            features=[guilds.GuildFeature.COMMUNITY, guilds.GuildFeature.RAID_ALERTS_DISABLED],
            reason="beep boop",
        )

        assert edited_guild is model.app.rest.edit_guild.return_value

    @pytest.mark.asyncio
    async def test_set_incident_actions(self, model: guilds.PartialGuild):
        model.app.rest.set_guild_incident_actions = mock.AsyncMock()

        updated_incident_data = await model.set_incident_actions(
            invites_disabled_until=datetime.datetime(2021, 11, 17), dms_disabled_until=datetime.datetime(2021, 11, 18)
        )

        assert updated_incident_data is model.app.rest.set_guild_incident_actions.return_value
        model.app.rest.set_guild_incident_actions.assert_awaited_once_with(
            90210,
            invites_disabled_until=datetime.datetime(2021, 11, 17),
            dms_disabled_until=datetime.datetime(2021, 11, 18),
        )

    @pytest.mark.asyncio
    async def test_fetch_emojis(self, model):
        model.app.rest.fetch_guild_emojis = mock.AsyncMock()

        emojis = await model.fetch_emojis()

        model.app.rest.fetch_guild_emojis.assert_awaited_once_with(model.id)
        assert emojis is model.app.rest.fetch_guild_emojis.return_value

    @pytest.mark.asyncio
    async def test_fetch_emoji(self, model):
        model.app.rest.fetch_emoji = mock.AsyncMock()

        emoji = await model.fetch_emoji(349)

        model.app.rest.fetch_emoji.assert_awaited_once_with(model.id, 349)
        assert emoji is model.app.rest.fetch_emoji.return_value

    @pytest.mark.asyncio
    async def test_fetch_stickers(self, model):
        model.app.rest.fetch_guild_stickers = mock.AsyncMock()

        stickers = await model.fetch_stickers()

        model.app.rest.fetch_guild_stickers.assert_awaited_once_with(model.id)
        assert stickers is model.app.rest.fetch_guild_stickers.return_value

    @pytest.mark.asyncio
    async def test_fetch_sticker(self, model):
        model.app.rest.fetch_guild_sticker = mock.AsyncMock()

        sticker = await model.fetch_sticker(6969)

        model.app.rest.fetch_guild_sticker.assert_awaited_once_with(model.id, 6969)
        assert sticker is model.app.rest.fetch_guild_sticker.return_value

    @pytest.mark.asyncio
    async def test_create_sticker(self, model):
        model.app.rest.create_sticker = mock.AsyncMock()
        file = object()

        sticker = await model.create_sticker(
            "NewSticker", "funny", file, description="A sticker", reason="blah blah blah"
        )
        assert sticker is model.app.rest.create_sticker.return_value

        model.app.rest.create_sticker.assert_awaited_once_with(
            90210, "NewSticker", "funny", file, description="A sticker", reason="blah blah blah"
        )

    @pytest.mark.asyncio
    async def test_edit_sticker(self, model):
        model.app.rest.edit_sticker = mock.AsyncMock()

        sticker = await model.edit_sticker(4567, name="Brilliant", tag="parmesan", description="amazing")

        model.app.rest.edit_sticker.assert_awaited_once_with(
            90210, 4567, name="Brilliant", tag="parmesan", description="amazing", reason=undefined.UNDEFINED
        )

        assert sticker is model.app.rest.edit_sticker.return_value

    @pytest.mark.asyncio
    async def test_delete_sticker(self, model):
        model.app.rest.delete_sticker = mock.AsyncMock()

        sticker = await model.delete_sticker(951)

        model.app.rest.delete_sticker.assert_awaited_once_with(90210, 951, reason=undefined.UNDEFINED)

        assert sticker is model.app.rest.delete_sticker.return_value

    @pytest.mark.asyncio
    async def test_create_category(self, model):
        model.app.rest.create_guild_category = mock.AsyncMock()

        category = await model.create_category("very cool category", position=2)

        model.app.rest.create_guild_category.assert_awaited_once_with(
            90210,
            "very cool category",
            position=2,
            permission_overwrites=undefined.UNDEFINED,
            reason=undefined.UNDEFINED,
        )

        assert category is model.app.rest.create_guild_category.return_value

    @pytest.mark.asyncio
    async def test_create_text_channel(self, model):
        model.app.rest.create_guild_text_channel = mock.AsyncMock()

        text_channel = await model.create_text_channel(
            "cool text channel", position=3, nsfw=False, rate_limit_per_user=30
        )

        model.app.rest.create_guild_text_channel.assert_awaited_once_with(
            90210,
            "cool text channel",
            position=3,
            topic=undefined.UNDEFINED,
            nsfw=False,
            rate_limit_per_user=30,
            permission_overwrites=undefined.UNDEFINED,
            category=undefined.UNDEFINED,
            reason=undefined.UNDEFINED,
        )

        assert text_channel is model.app.rest.create_guild_text_channel.return_value

    @pytest.mark.asyncio
    async def test_create_news_channel(self, model):
        model.app.rest.create_guild_news_channel = mock.AsyncMock()

        news_channel = await model.create_news_channel(
            "cool news channel", position=1, nsfw=False, rate_limit_per_user=420
        )

        model.app.rest.create_guild_news_channel.assert_awaited_once_with(
            90210,
            "cool news channel",
            position=1,
            topic=undefined.UNDEFINED,
            nsfw=False,
            rate_limit_per_user=420,
            permission_overwrites=undefined.UNDEFINED,
            category=undefined.UNDEFINED,
            reason=undefined.UNDEFINED,
        )

        assert news_channel is model.app.rest.create_guild_news_channel.return_value

    @pytest.mark.asyncio
    async def test_create_forum_channel(self, model):
        model.app.rest.create_guild_forum_channel = mock.AsyncMock()

        forum_channel = await model.create_forum_channel(
            "cool forum channel", position=1, nsfw=False, rate_limit_per_user=420
        )

        model.app.rest.create_guild_forum_channel.assert_awaited_once_with(
            90210,
            "cool forum channel",
            position=1,
            topic=undefined.UNDEFINED,
            nsfw=False,
            rate_limit_per_user=420,
            permission_overwrites=undefined.UNDEFINED,
            category=undefined.UNDEFINED,
            reason=undefined.UNDEFINED,
            default_auto_archive_duration=undefined.UNDEFINED,
            default_thread_rate_limit_per_user=undefined.UNDEFINED,
            default_forum_layout=undefined.UNDEFINED,
            default_sort_order=undefined.UNDEFINED,
            available_tags=undefined.UNDEFINED,
            default_reaction_emoji=undefined.UNDEFINED,
        )

        assert forum_channel is model.app.rest.create_guild_forum_channel.return_value

    @pytest.mark.asyncio
    async def test_create_voice_channel(self, model):
        model.app.rest.create_guild_voice_channel = mock.AsyncMock()

        voice_channel = await model.create_voice_channel(
            "cool voice channel", position=1, bitrate=3200, video_quality_mode=2
        )

        model.app.rest.create_guild_voice_channel.assert_awaited_once_with(
            90210,
            "cool voice channel",
            position=1,
            user_limit=undefined.UNDEFINED,
            bitrate=3200,
            video_quality_mode=2,
            permission_overwrites=undefined.UNDEFINED,
            region=undefined.UNDEFINED,
            category=undefined.UNDEFINED,
            reason=undefined.UNDEFINED,
        )

        assert voice_channel is model.app.rest.create_guild_voice_channel.return_value

    @pytest.mark.asyncio
    async def test_create_stage_channel(self, model):
        model.app.rest.create_guild_stage_channel = mock.AsyncMock()

        stage_channel = await model.create_stage_channel("cool stage channel", position=1, bitrate=3200, user_limit=100)

        model.app.rest.create_guild_stage_channel.assert_awaited_once_with(
            90210,
            "cool stage channel",
            position=1,
            user_limit=100,
            bitrate=3200,
            permission_overwrites=undefined.UNDEFINED,
            region=undefined.UNDEFINED,
            category=undefined.UNDEFINED,
            reason=undefined.UNDEFINED,
        )

        assert stage_channel is model.app.rest.create_guild_stage_channel.return_value

    @pytest.mark.asyncio
    async def test_delete_channel(self, model):
        mock_channel = mock.Mock(channels_.GuildChannel)
        model.app.rest.delete_channel = mock.AsyncMock(return_value=mock_channel)

        deleted_channel = await model.delete_channel(1288820)

        model.app.rest.delete_channel.assert_awaited_once_with(1288820)
        assert deleted_channel is model.app.rest.delete_channel.return_value

    @pytest.mark.asyncio
    async def test_fetch_guild(self, model):
        model.app.rest.fetch_guild = mock.AsyncMock(return_value=model)

        assert await model.fetch_self() is model.app.rest.fetch_guild.return_value
        model.app.rest.fetch_guild.assert_awaited_once_with(model.id)

    @pytest.mark.asyncio
    async def test_fetch_roles(self, model):
        model.app.rest.fetch_roles = mock.AsyncMock()

        roles = await model.fetch_roles()

        model.app.rest.fetch_roles.assert_awaited_once_with(90210)
        assert roles is model.app.rest.fetch_roles.return_value


class TestGuildPreview:
    @pytest.fixture
    def model(self, mock_app):
        return guilds.GuildPreview(
            app=mock_app,
            features=["huge super secret nsfw channel"],
            id=snowflakes.Snowflake(123),
            icon_hash="dis is mah icon hash",
            name="DAPI",
            splash_hash="dis is also mah splash hash",
            discovery_splash_hash=None,
            emojis={},
            approximate_active_member_count=12,
            approximate_member_count=999_283_252_124_633,
            description="the place for quality shitposting!",
        )

    def test_make_splash_url_format_set_to_deprecated_ext_argument_if_provided(self, model):
        with mock.patch.object(
            routes, "CDN_GUILD_SPLASH", new=mock.Mock(compile_to_file=mock.Mock(return_value="file"))
        ) as route:
            assert model.make_splash_url(ext="JPEG") == "file"

        route.compile_to_file.assert_called_once_with(
            urls.CDN_URL, guild_id=123, hash="dis is also mah splash hash", size=4096, file_format="JPEG", lossless=True
        )

    def test_make_discovery_splash_url_format_set_to_deprecated_ext_argument_if_provided(self, model):
        model.discovery_splash_hash = "18dnf8dfbakfdh"

        with mock.patch.object(
            routes, "CDN_GUILD_DISCOVERY_SPLASH", new=mock.Mock(compile_to_file=mock.Mock(return_value="file"))
        ) as route:
            assert model.make_discovery_splash_url(ext="JPEG") == "file"

        route.compile_to_file.assert_called_once_with(
            urls.CDN_URL, guild_id=123, hash="18dnf8dfbakfdh", size=4096, file_format="JPEG", lossless=True
        )

    def test_splash_url(self, model):
        splash = object()

        with mock.patch.object(guilds.GuildPreview, "make_splash_url", return_value=splash):
            assert model.splash_url is splash

    def test_make_splash_url_when_hash(self, model):
        model.splash_hash = "18dnf8dfbakfdh"

        with mock.patch.object(
            routes, "CDN_GUILD_SPLASH", new=mock.Mock(compile_to_file=mock.Mock(return_value="file"))
        ) as route:
            assert model.make_splash_url(file_format="URL", size=1024) == "file"

        route.compile_to_file.assert_called_once_with(
            urls.CDN_URL, guild_id=123, hash="18dnf8dfbakfdh", size=1024, file_format="URL", lossless=True
        )

    def test_make_splash_url_when_no_hash(self, model):
        model.splash_hash = None
        assert model.make_splash_url(ext="png", size=512) is None

    def test_discovery_splash_url(self, model):
        discovery_splash = object()

        with mock.patch.object(guilds.GuildPreview, "make_discovery_splash_url", return_value=discovery_splash):
            assert model.discovery_splash_url is discovery_splash

    def test_make_discovery_splash_url_when_hash(self, model):
        model.discovery_splash_hash = "18dnf8dfbakfdh"

        with mock.patch.object(
            routes, "CDN_GUILD_DISCOVERY_SPLASH", new=mock.Mock(compile_to_file=mock.Mock(return_value="file"))
        ) as route:
            assert model.make_discovery_splash_url(file_format="URL", size=2048) == "file"

        route.compile_to_file.assert_called_once_with(
            urls.CDN_URL, guild_id=123, hash="18dnf8dfbakfdh", size=2048, file_format="URL", lossless=True
        )

    def test_make_discovery_splash_url_when_no_hash(self, model):
        model.discovery_splash_hash = None
        assert model.make_discovery_splash_url(file_format="PNG", size=4096) is None


class TestGuild:
    @pytest.fixture
    def model(self, mock_app):
        return hikari_test_helpers.mock_class_namespace(guilds.Guild)(
            app=mock_app,
            id=snowflakes.Snowflake(123),
            splash_hash="splash_hash",
            discovery_splash_hash="discovery_splash_hash",
            banner_hash="banner_hash",
            icon_hash="icon_hash",
            features=[guilds.GuildFeature.ANIMATED_ICON],
            incidents=guilds.GuildIncidents(
                invites_disabled_until=None,
                dms_disabled_until=None,
                dm_spam_detected_at=datetime.datetime(2015, 5, 13, 1, 1, 1, 1, tzinfo=datetime.timezone.utc),
                raid_detected_at=None,
            ),
            name="some guild",
            application_id=snowflakes.Snowflake(9876),
            afk_channel_id=snowflakes.Snowflake(1234),
            afk_timeout=datetime.timedelta(seconds=60),
            default_message_notifications=guilds.GuildMessageNotificationsLevel.ONLY_MENTIONS,
            description=None,
            explicit_content_filter=guilds.GuildExplicitContentFilterLevel.ALL_MEMBERS,
            is_widget_enabled=False,
            max_video_channel_users=10,
            mfa_level=guilds.GuildMFALevel.NONE,
            nsfw_level=guilds.GuildNSFWLevel.AGE_RESTRICTED,
            owner_id=snowflakes.Snowflake(1111),
            preferred_locale="en-GB",
            premium_subscription_count=12,
            premium_tier=guilds.GuildPremiumTier.TIER_3,
            public_updates_channel_id=snowflakes.Snowflake(99699),
            rules_channel_id=snowflakes.Snowflake(123445),
            system_channel_id=snowflakes.Snowflake(123888),
            vanity_url_code="yeet",
            verification_level=guilds.GuildVerificationLevel.VERY_HIGH,
            widget_channel_id=snowflakes.Snowflake(192729),
            system_channel_flags=guilds.GuildSystemChannelFlag.SUPPRESS_PREMIUM_SUBSCRIPTION,
        )

    def test_get_channels(self, model):
        assert model.get_channels() is model.app.cache.get_guild_channels_view_for_guild.return_value
        model.app.cache.get_guild_channels_view_for_guild.assert_called_once_with(123)

    def test_get_channels_when_no_cache_trait(self, model):
        model.app = object()
        assert model.get_channels() == {}

    def test_get_members(self, model):
        assert model.get_members() is model.app.cache.get_members_view_for_guild.return_value
        model.app.cache.get_members_view_for_guild.assert_called_once_with(123)

    def test_get_members_when_no_cache_trait(self, model):
        model.app = object()
        assert model.get_members() == {}

    def test_get_presences(self, model):
        assert model.get_presences() is model.app.cache.get_presences_view_for_guild.return_value
        model.app.cache.get_presences_view_for_guild.assert_called_once_with(123)

    def test_get_presences_when_no_cache_trait(self, model):
        model.app = object()
        assert model.get_presences() == {}

    def test_get_voice_states(self, model):
        assert model.get_voice_states() is model.app.cache.get_voice_states_view_for_guild.return_value
        model.app.cache.get_voice_states_view_for_guild.assert_called_once_with(123)

    def test_get_voice_states_when_no_cache_trait(self, model):
        model.app = object()
        assert model.get_voice_states() == {}

    def test_get_emojis(self, model):
        assert model.get_emojis() is model.app.cache.get_emojis_view_for_guild.return_value
        model.app.cache.get_emojis_view_for_guild.assert_called_once_with(123)

    def test_emojis_when_no_cache_trait(self, model):
        model.app = object()
        assert model.get_emojis() == {}

    def test_get_sticker(self, model):
        model.app.cache.get_sticker.return_value.guild_id = model.id
        assert model.get_sticker(456) is model.app.cache.get_sticker.return_value
        model.app.cache.get_sticker.assert_called_once_with(456)

    def test_get_sticker_when_not_from_guild(self, model):
        model.app.cache.get_sticker.return_value.guild_id = 546123123433

        assert model.get_sticker(456) is None

        model.app.cache.get_sticker.assert_called_once_with(456)

    def test_get_sticker_when_no_cache_trait(self, model):
        model.app = object()
        assert model.get_sticker(1234) is None

    def test_get_stickers(self, model):
        assert model.get_stickers() is model.app.cache.get_stickers_view_for_guild.return_value
        model.app.cache.get_stickers_view_for_guild.assert_called_once_with(123)

    def test_get_stickers_when_no_cache_trait(self, model):
        model.app = object()
        assert model.get_stickers() == {}

    def test_roles(self, model):
        assert model.get_roles() is model.app.cache.get_roles_view_for_guild.return_value
        model.app.cache.get_roles_view_for_guild.assert_called_once_with(123)

    def test_get_roles_when_no_cache_trait(self, model):
        model.app = object()
        assert model.get_roles() == {}

    def test_get_emoji(self, model):
        model.app.cache.get_emoji.return_value.guild_id = model.id
        assert model.get_emoji(456) is model.app.cache.get_emoji.return_value
        model.app.cache.get_emoji.assert_called_once_with(456)

    def test_get_emoji_when_not_from_guild(self, model):
        model.app.cache.get_emoji.return_value.guild_id = 1233212

        assert model.get_emoji(456) is None

        model.app.cache.get_emoji.assert_called_once_with(456)

    def test_get_emoji_when_no_cache_trait(self, model):
        model.app = object()
        assert model.get_emoji(456) is None

    def test_get_role(self, model):
        model.app.cache.get_role.return_value.guild_id = model.id
        assert model.get_role(456) is model.app.cache.get_role.return_value
        model.app.cache.get_role.assert_called_once_with(456)

    def test_get_role_when_not_from_guild(self, model):
        model.app.cache.get_role.return_value.guild_id = 7623123321123

        assert model.get_role(456) is None

        model.app.cache.get_role.assert_called_once_with(456)

    def test_get_role_when_no_cache_trait(self, model):
        model.app = object()
        assert model.get_role(456) is None

    @pytest.mark.asyncio
    async def test_invites_disabled_default(self, model):
        assert model.invites_disabled is False

    @pytest.mark.asyncio
    async def test_invites_disabled_via_incidents(self, model):
        model.incidents = guilds.GuildIncidents(
            invites_disabled_until=datetime.datetime(2021, 11, 17),
            dms_disabled_until=None,
            dm_spam_detected_at=None,
            raid_detected_at=None,
        )
        assert model.invites_disabled is True

    @pytest.mark.asyncio
    async def test_invites_disabled_via_feature(self, model):
        model.features.append(guilds.GuildFeature.INVITES_DISABLED)
        assert model.invites_disabled is True

    def test_make_banner_url_format_set_to_deprecated_ext_argument_if_provided(self, model):
        with mock.patch.object(
            routes, "CDN_GUILD_BANNER", new=mock.Mock(compile_to_file=mock.Mock(return_value="file"))
        ) as route:
            assert model.make_banner_url(ext="JPEG") == "file"

        route.compile_to_file.assert_called_once_with(
            urls.CDN_URL, guild_id=123, hash="banner_hash", size=4096, file_format="JPEG", lossless=True
        )

    def test_make_discovery_splash_url_format_set_to_deprecated_ext_argument_if_provided(self, model):
        with mock.patch.object(
            routes, "CDN_GUILD_DISCOVERY_SPLASH", new=mock.Mock(compile_to_file=mock.Mock(return_value="file"))
        ) as route:
            assert model.make_discovery_splash_url(ext="JPEG") == "file"

        route.compile_to_file.assert_called_once_with(
            urls.CDN_URL, guild_id=123, hash="discovery_splash_hash", size=4096, file_format="JPEG", lossless=True
        )

    def test_make_splash_url_format_set_to_deprecated_ext_argument_if_provided(self, model):
        with mock.patch.object(
            routes, "CDN_GUILD_SPLASH", new=mock.Mock(compile_to_file=mock.Mock(return_value="file"))
        ) as route:
            assert model.make_splash_url(ext="JPEG") == "file"

        route.compile_to_file.assert_called_once_with(
            urls.CDN_URL, guild_id=123, hash="splash_hash", size=4096, file_format="JPEG", lossless=True
        )

    def test_splash_url(self, model):
        splash = object()

        with mock.patch.object(guilds.Guild, "make_splash_url", return_value=splash):
            assert model.splash_url is splash

    def test_make_splash_url_when_hash(self, model):
        model.splash_hash = "18dnf8dfbakfdh"

        with mock.patch.object(
            routes, "CDN_GUILD_SPLASH", new=mock.Mock(compile_to_file=mock.Mock(return_value="file"))
        ) as route:
            assert model.make_splash_url(file_format="URL", size=2) == "file"

        route.compile_to_file.assert_called_once_with(
            urls.CDN_URL, guild_id=123, hash="18dnf8dfbakfdh", size=2, file_format="URL", lossless=True
        )

    def test_make_splash_url_when_no_hash(self, model):
        model.splash_hash = None
        assert model.make_splash_url(ext="png", size=1024) is None

    def test_discovery_splash_url(self, model):
        discovery_splash = object()

        with mock.patch.object(guilds.Guild, "make_discovery_splash_url", return_value=discovery_splash):
            assert model.discovery_splash_url is discovery_splash

    def test_make_discovery_splash_url_when_hash(self, model):
        model.discovery_splash_hash = "18dnf8dfbakfdh"

        with mock.patch.object(
            routes, "CDN_GUILD_DISCOVERY_SPLASH", new=mock.Mock(compile_to_file=mock.Mock(return_value="file"))
        ) as route:
            assert model.make_discovery_splash_url(file_format="URL", size=1024) == "file"

        route.compile_to_file.assert_called_once_with(
            urls.CDN_URL, guild_id=123, hash="18dnf8dfbakfdh", size=1024, file_format="URL", lossless=True
        )

    def test_make_discovery_splash_url_when_no_hash(self, model):
        model.discovery_splash_hash = None
        assert model.make_discovery_splash_url(ext="png", size=2048) is None

    def test_banner_url(self, model):
        banner = object()

        with mock.patch.object(guilds.Guild, "make_banner_url", return_value=banner):
            assert model.banner_url is banner

    def test_make_banner_url_when_hash(self, model):
        with mock.patch.object(
            routes, "CDN_GUILD_BANNER", new=mock.Mock(compile_to_file=mock.Mock(return_value="file"))
        ) as route:
            assert model.make_banner_url(file_format="URL", size=512) == "file"

        route.compile_to_file.assert_called_once_with(
            urls.CDN_URL, guild_id=123, hash="banner_hash", size=512, file_format="URL", lossless=True
        )

    def test_make_banner_url_when_format_is_None_and_banner_hash_is_for_animated(self, model):
        model.banner_hash = "a_18dnf8dfbakfdh"

        with mock.patch.object(
            routes, "CDN_GUILD_BANNER", new=mock.Mock(compile_to_file=mock.Mock(return_value="file"))
        ) as route:
            assert model.make_banner_url(file_format=None, size=4096) == "file"

        route.compile_to_file.assert_called_once_with(
            urls.CDN_URL, guild_id=model.id, hash="a_18dnf8dfbakfdh", size=4096, file_format="GIF", lossless=True
        )

    def test_make_banner_url_when_format_is_None_and_banner_hash_is_not_for_animated(self, model):
        model.banner_hash = "18dnf8dfbakfdh"

        with mock.patch.object(
            routes, "CDN_GUILD_BANNER", new=mock.Mock(compile_to_file=mock.Mock(return_value="file"))
        ) as route:
            assert model.make_banner_url(file_format=None, size=4096) == "file"

        route.compile_to_file.assert_called_once_with(
            urls.CDN_URL, guild_id=model.id, hash=model.banner_hash, size=4096, file_format="PNG", lossless=True
        )

    def test_make_banner_url_when_no_hash(self, model):
        model.banner_hash = None
        assert model.make_banner_url(file_format="PNG", size=2048) is None

    @pytest.mark.asyncio
    async def test_fetch_owner(self, model):
        model.app.rest.fetch_member = mock.AsyncMock()

        assert await model.fetch_owner() is model.app.rest.fetch_member.return_value
        model.app.rest.fetch_member.assert_awaited_once_with(123, 1111)

    @pytest.mark.asyncio
    async def test_fetch_widget_channel(self, model):
        mock_channel = mock.Mock(channels_.GuildChannel)
        model.app.rest.fetch_channel = mock.AsyncMock(return_value=mock_channel)

        assert await model.fetch_widget_channel() is model.app.rest.fetch_channel.return_value
        model.app.rest.fetch_channel.assert_awaited_once_with(192729)

    @pytest.mark.asyncio
    async def test_fetch_widget_channel_when_None(self, model):
        model.widget_channel_id = None

        assert await model.fetch_widget_channel() is None

    @pytest.mark.asyncio
    async def test_fetch_rules_channel(self, model):
        mock_channel = mock.Mock(channels_.GuildTextChannel)
        model.app.rest.fetch_channel = mock.AsyncMock(return_value=mock_channel)

        assert await model.fetch_rules_channel() is model.app.rest.fetch_channel.return_value
        model.app.rest.fetch_channel.assert_awaited_once_with(123445)

    @pytest.mark.asyncio
    async def test_fetch_rules_channel_when_None(self, model):
        model.rules_channel_id = None

        assert await model.fetch_rules_channel() is None

    @pytest.mark.asyncio
    async def test_fetch_system_channel(self, model):
        mock_channel = mock.Mock(channels_.GuildTextChannel)
        model.app.rest.fetch_channel = mock.AsyncMock(return_value=mock_channel)

        assert await model.fetch_system_channel() is model.app.rest.fetch_channel.return_value
        model.app.rest.fetch_channel.assert_awaited_once_with(123888)

    @pytest.mark.asyncio
    async def test_fetch_system_channel_when_None(self, model):
        model.system_channel_id = None

        assert await model.fetch_system_channel() is None

    @pytest.mark.asyncio
    async def test_fetch_public_updates_channel(self, model):
        mock_channel = mock.Mock(channels_.GuildTextChannel)
        model.app.rest.fetch_channel = mock.AsyncMock(return_value=mock_channel)

        assert await model.fetch_public_updates_channel() is model.app.rest.fetch_channel.return_value
        model.app.rest.fetch_channel.assert_awaited_once_with(99699)

    @pytest.mark.asyncio
    async def test_fetch_public_updates_channel_when_None(self, model):
        model.public_updates_channel_id = None

        assert await model.fetch_public_updates_channel() is None

    @pytest.mark.asyncio
    async def test_fetch_afk_channel(self, model):
        mock_channel = mock.Mock(channels_.GuildVoiceChannel)
        model.app.rest.fetch_channel = mock.AsyncMock(return_value=mock_channel)

        assert await model.fetch_afk_channel() is model.app.rest.fetch_channel.return_value
        model.app.rest.fetch_channel.assert_awaited_once_with(1234)

    @pytest.mark.asyncio
    async def test_fetch_afk_channel_when_None(self, model):
        model.afk_channel_id = None

        assert await model.fetch_afk_channel() is None

    def test_get_channel(self, model):
        model.app.cache.get_guild_channel.return_value.guild_id = model.id
        assert model.get_channel(456) is model.app.cache.get_guild_channel.return_value
        model.app.cache.get_guild_channel.assert_called_once_with(456)

    def test_get_channel_when_not_from_guild(self, model):
        model.app.cache.get_guild_channel.return_value.guild_id = 654523123

        assert model.get_channel(456) is None

        model.app.cache.get_guild_channel.assert_called_once_with(456)

    def test_get_channel_when_no_cache_trait(self, model):
        model.app = object()
        assert model.get_channel(456) is None

    def test_get_member(self, model):
        assert model.get_member(456) is model.app.cache.get_member.return_value
        model.app.cache.get_member.assert_called_once_with(123, 456)

    def test_get_member_when_no_cache_trait(self, model):
        model.app = object()
        assert model.get_member(456) is None

    def test_get_presence(self, model):
        assert model.get_presence(456) is model.app.cache.get_presence.return_value
        model.app.cache.get_presence.assert_called_once_with(123, 456)

    def test_get_presence_when_no_cache_trait(self, model):
        model.app = object()
        assert model.get_presence(456) is None

    def test_get_voice_state(self, model):
        assert model.get_voice_state(456) is model.app.cache.get_voice_state.return_value
        model.app.cache.get_voice_state.assert_called_once_with(123, 456)

    def test_get_voice_state_when_no_cache_trait(self, model):
        model.app = object()
        assert model.get_voice_state(456) is None

    def test_get_my_member_when_not_shardaware(self, model):
        model.app = object()
        assert model.get_my_member() is None

    def test_get_my_member_when_no_me(self, model):
        model.app.get_me = mock.Mock(return_value=None)

        assert model.get_my_member() is None

        model.app.get_me.assert_called_once_with()

    def test_get_my_member(self, model):
        model.app.get_me = mock.Mock()
        model.app.get_me.return_value.id = 123

        with mock.patch.object(guilds.Guild, "get_member") as get_member:
            assert model.get_my_member() is get_member.return_value

        get_member.assert_called_once_with(123)
        model.app.get_me.assert_called_once_with()


class TestRestGuild:
    @pytest.fixture
    def model(self, mock_app):
        return guilds.RESTGuild(
            app=mock_app,
            id=snowflakes.Snowflake(123),
            splash_hash="splash_hash",
            discovery_splash_hash="discovery_splash_hash",
            banner_hash="banner_hash",
            icon_hash="icon_hash",
            features=[guilds.GuildFeature.ANIMATED_ICON],
            incidents=guilds.GuildIncidents(
                invites_disabled_until=None,
                dms_disabled_until=None,
                dm_spam_detected_at=datetime.datetime(2015, 5, 13, 1, 1, 1, 1, tzinfo=datetime.timezone.utc),
                raid_detected_at=None,
            ),
            name="some guild",
            application_id=snowflakes.Snowflake(9876),
            afk_channel_id=snowflakes.Snowflake(1234),
            afk_timeout=datetime.timedelta(seconds=60),
            default_message_notifications=guilds.GuildMessageNotificationsLevel.ONLY_MENTIONS,
            description=None,
            explicit_content_filter=guilds.GuildExplicitContentFilterLevel.ALL_MEMBERS,
            is_widget_enabled=False,
            max_video_channel_users=10,
            mfa_level=guilds.GuildMFALevel.NONE,
            owner_id=snowflakes.Snowflake(1111),
            preferred_locale="en-GB",
            premium_subscription_count=12,
            premium_tier=guilds.GuildPremiumTier.TIER_3,
            public_updates_channel_id=None,
            rules_channel_id=None,
            system_channel_id=None,
            vanity_url_code="yeet",
            verification_level=guilds.GuildVerificationLevel.VERY_HIGH,
            widget_channel_id=None,
            system_channel_flags=guilds.GuildSystemChannelFlag.SUPPRESS_PREMIUM_SUBSCRIPTION,
            emojis={},
            stickers={},
            roles={},
            approximate_active_member_count=1000,
            approximate_member_count=100,
            max_presences=100,
            max_members=100,
            nsfw_level=guilds.GuildNSFWLevel.AGE_RESTRICTED,
        )
